from rich.console import Console


richconsole: Console = Console(highlight=False)
success_stm: str = '[green]Success![/]'
info_stm: str = '[blue1]Info![/]'
warn_stm: str = '[red]Warning![/]'
error_stm: str = '[red]Error![/]'
debug_stm: str = '[orange1]Debug![/]'
critical_stm: str = '[red]Critical![/]'

__version__ = 'v1.0.0'


__all__ = [
    "richconsole",
    "success_stm",
    "info_stm",
    "warn_stm",
    "error_stm",
    "debug_stm",
    "critical_stm",
    "__version__"]



from montyalex.fs_tools import cancel
from montyalex.object_tools import (
    Key,
    Value,
    MontyKeyValueError as KeyValError,
    MontyJSONError as JSONDecoderError,
    MontyRangeError as RangeError)
from montyalex.typing_tools import Any, Literal
from montyalex.uo_tools import yaml


class State:
    def __init__(
        self,
        file_name: str,
        *,
        multiplier: int = 1,
        limit: int | tuple[int, int] = None,
        ilimit: int | tuple[int, int] = None,
        loc: str = '.mtax') -> None:
        __yaml = yaml(directory=loc, filename=file_name)
        self.multiplier: int = multiplier
        self.limit: int | tuple[int, int] = limit
        self.ilimit: int | tuple[int, int] = ilimit
        self.file_name: str = file_name
        self.file_path: str = f'{file_name}.{__yaml.modelname}'
        self.full_path: str = __yaml.modelpath
        self.loc: str = loc
        self.yamlchange = __yaml.change
        self.yamlexists = __yaml.exists
        self.yamlmodel = __yaml.model
        self.yamlremove = __yaml.remove

    def get(self, key: Key):
        key_value = self.get_value(key)
        if key_value is not None:
            state: dict[Key, Any] = {key: key_value}
            print(f'Key Found {state}')
        else:
            raise KeyValError('No value associated')

    def get_value(self, key: Key):
        try:
            key_value = self.yamlmodel(read=True)[key]
            if key_value is not None:
                return key_value
            raise KeyValError(f'Null assignment for {key!r}; no value associated with active state')
        except (KeyError, TypeError, JSONDecoderError):
            print(f"KeyError: {key!r} can not be found in file; no value was assigned")
            return None

    def delete(self, key: Key = None, value: Value = None, *, fulloc: bool = False):
        if key is None and value is None:
            path = self.file_path
            if fulloc:
                path = self.full_path
            if self.yamlexists:
                self.yamlremove()
                print(f"Removed {path!r}")
            else:
                raise FileNotFoundError(f'the state file {path!r} is missing')
        if key and not value:
            key_value = self.get_value(key)
            state: dict[Key, Any] = {key: key_value}
            if key_value is not None:
                print(f"Removed {state} from {self.file_name!r}")
                self.yamlchange(key, None)
            else:
                raise KeyValError(
                    f"{key!r} was not nullified {state}, no value exists for defined key")
        if key and value:
            key_value = self.get_value(key)
            state: dict[Key, Any] = {key: value}
            if key_value == value:
                print(f"Removed {state} from {self.file_name!r}")
                self.yamlchange(key, None)
            else:
                raise KeyValError(
                    f"{key!r} was not nullified {state}, value ({value!r}) does not match")

    def manager(self, key: Key, value: Value):
        state: dict[Key, Any] = {key: value}
        print(f"State Changed {state}")
        if self.yamlexists and self.get_value(key):
            self.yamlmodel(state, write=True, append=True)
        else:
            self.yamlmodel(state, write=True)

    def _modifier(
        self,
        value: Value,
        operation: Literal[0] | Literal[1] | Literal[2] | Literal[3],
        *,
        batch: int):
        check = value
        if value is not None:
            check = value + self.multiplier if operation == 0 else value - self.multiplier
            if operation == 2:
                check = value + batch
            if operation == 3:
                check = value - batch
            if check % batch != 0:
                check = f'{check}, batch % {batch} != 0'
        if check is None:
            check = f'±0, {{{operation}, {value}*{self.multiplier}}}'
        if isinstance(self.limit, tuple) or isinstance(self.ilimit, tuple):
            inclusive_limit = range(self.ilimit[0], self.ilimit[1] + 1)
            range_limit = range(*self.limit) if not self.ilimit else inclusive_limit
            if value not in range_limit or check not in range_limit:
                if self.limit:
                    raise RangeError(f'Index ({check}) not in', range(*self.limit))
                if self.ilimit:
                    raise RangeError(f'Index ({check}) not in', range(*self.ilimit))
                cancel()
        if isinstance(self.limit, int) or isinstance(self.ilimit, int):
            range_limit = range(self.limit) if not self.ilimit else range(self.ilimit + 1)
            if value not in range_limit or check not in range_limit:
                if self.limit:
                    raise RangeError(f'Index ({check}) not in', range(self.limit))
                if self.ilimit:
                    raise RangeError(f'Index ({check}) not in', range(self.ilimit))
                cancel()
        return True

    def decr(self, key: Key):
        key_value = self.get_value(key)
        old_state: dict[Key, Any] = {key: key_value}
        state: dict[Key, Any] = {}
        if isinstance(key_value, int):
            state: dict[Key, Any] = {key: key_value - self.multiplier}
        try:
            smdecr = self._modifier(key_value, 1, batch=self.multiplier)
            if state[key] != key_value and smdecr:
                print(f"Decremented {old_state} -> {state}")
                self.yamlmodel(state, write=True, append=True)
        except (KeyValError, RangeError) as e:
            print(e)

    def incr(self, key: Key):
        key_value = self.get_value(key)
        old_state: dict[Key, Any] = {key: key_value}
        state: dict[Key, Any] = {}
        if isinstance(key_value, int):
            state: dict[Key, Any] = {key: key_value + self.multiplier}
        try:
            smincr = self._modifier(key_value, 0, batch=self.multiplier)
            if state[key] != key_value and smincr:
                print(f"Incremented {old_state} -> {state}")
                self.yamlmodel(state, write=True, append=True)
        except (KeyValError, RangeError) as e:
            print(e)

    def bypass(self, key: Key, batch: int):
        key_value = self.get_value(key)
        old_state: dict[Key, Any] = {key: key_value}
        state: dict[Key, Any] = {}
        if isinstance(key_value, int):
            state: dict[Key, Any] = {key: key_value + batch}
        try:
            smbatch = self._modifier(key_value, 2, batch=batch)
            if state[key] != key_value and smbatch:
                print(f"Incremented {old_state} -> (batch={batch}) -> {state}")
                self.yamlmodel(state, write=True, append=True)
        except (KeyValError, RangeError) as e:
            print(e)

    def nbypass(self, key: Key, batch: int):
        key_value = self.get_value(key)
        old_state: dict[Key, Any] = {key: key_value}
        state: dict[Key, Any] = {}
        if isinstance(key_value, int):
            state: dict[Key, Any] = {key: key_value - batch}
        try:
            smbatch = self._modifier(key_value, 3, batch=batch)
            if state[key] != key_value and smbatch:
                print(f"Decremented {old_state} -> (batch={batch}) -> {state}")
                self.yamlmodel(state, write=True, append=True)
        except (KeyValError, RangeError) as e:
            print(e)


# ---------------------------------------------------------------------------
# | Example Usage
# ---------------------------------------------------------------------------
if __name__ == '__main__':
    class New:
        def __init__(self, file_name) -> None:
            self.state = State(file_name)
    mystate = New('test-state').state
    mystate.ilimit = (-100, 100)
    # mystate.manager('index', 0)
    mystate.nbypass('index', 10)
    mystate.nbypass('index', 10)
    mystate.nbypass('index', 10)
    # state.delete()



from montyalex.console_tools import richconsole, success_stm
from montyalex.fs_tools import current_working_dir, joinpaths, mkdirs, pathexists
from .helpdirs import remove_formated__dir

print = richconsole.print


def create_simple_directories(
    range_: int = 1,
    name_: str = None,
    prefix: str = None,
    suffix: str = None,
    silent: bool = False):
    for i in range(range_):
        formatted_directory_name = name_ if name_ else f'{i:03}'
        if prefix:
            formatted_directory_name = f'{prefix}{formatted_directory_name}'
        if suffix:
            formatted_directory_name = f'{formatted_directory_name}{suffix}'
        mkdirs(joinpaths(current_working_dir, formatted_directory_name), exist_ok=True)
        if not silent:
            print(
                f'{success_stm}, Created {formatted_directory_name!r} in {current_working_dir!r}')

def remove_simple_directories(
    range_: int = 1,
    name_: str = None,
    prefix: str = None,
    suffix: str = None,
    silent: bool = False):
    for i in range(range_):
        formatted_directory_name = name_ if name_ else f'{i:03}'
        if prefix:
            formatted_directory_name = f'{prefix}{formatted_directory_name}'
        if suffix:
            formatted_directory_name = f'{formatted_directory_name}{suffix}'
        formatted_directory = joinpaths(current_working_dir, formatted_directory_name)
        remove_formated__dir(formatted_directory, silent)

        if not pathexists(formatted_directory) and not silent:
            print(
                f'{success_stm}, Removed {formatted_directory_name!r} from {current_working_dir!r}')



from .cache_tools import (
    MtaxCache,
    add_to_cache,
    set_item_to_cache,
    get_item_from_cache,
    clear_cache,
    remove_item_from_cache,
    list_items_in_cache,
    show_info_of_cache,
    incr_in_cache,
    decr_in_cache)
from .console_tools import (
    richconsole,
    success_stm,
    info_stm,
    warn_stm,
    error_stm,
    debug_stm,
    critical_stm)
from .directory_tools import (
    datedirs, rmdatedirs, simpledirs, rmsimpledirs
)
from .future_tools import __november24__, __december24__
from .time_tools import MtaxTime


__all__ = [
    "MtaxCache",
    "add_to_cache",
    "set_item_to_cache",
    "get_item_from_cache",
    "clear_cache",
    "remove_item_from_cache",
    "list_items_in_cache",
    "show_info_of_cache",
    "incr_in_cache",
    "decr_in_cache",
    "richconsole",
    "success_stm",
    "info_stm",
    "warn_stm",
    "error_stm",
    "debug_stm",
    "critical_stm",
    "datedirs",
    "rmdatedirs",
    "simpledirs",
    "rmsimpledirs",
    "__november24__",
    "__december24__",
    "MtaxTime"]



from typing import Any, Callable, Dict, Generator, Generic, Literal, Optional, Type, TypeVar, Union
from types import NoneType


__version__ = 'v1.0.0'


__all__ = [
    "Any",
    "Callable",
    "Dict",
    "Generic",
    "Generator",
    "NoneType",
    "Literal",
    "Optional",
    "Type",
    "TypeVar",
    "Union",
    "__version__"]



from montyalex.time_tools import MtaxTime
from montyalex.typer_tools import Option, Typer
from .mtax_cache import MtaxCache
from .zonecache_app import zone_


time_: Typer = Typer(
    name='time',
    help='Manage time-data in the cache')
time_.add_typer(zone_)

@time_.command(name='reset')
def reset_(
    *,
    timezone: str = 'Etc/Greenwich',
    silent: bool = Option(
    False, '-s', show_default='-!s', help='Silence extreneous messages in console')):
    """Reset timestamp in the cache"""
    mtax: MtaxCache = MtaxCache(timezone)
    mtax.remove('_timestamp_', silent)
    set_(timezone=timezone, silent=silent)

@time_.command(name='set')
def set_(
    *,
    timezone: str = 'Etc/Greenwich',
    silent: bool = Option(
    False, '-s', show_default='-!s', help='Silence extreneous messages in console')):
    """Set timestamp in the cache"""
    mtax: MtaxCache = MtaxCache(timezone)
    mtax.set_item(('_timestamp_', MtaxTime(timezone).timestamp()), silent)

@time_.command(name='get')
def get_(
    *,
    timezone: str = 'Etc/Greenwich',
    silent: bool = Option(
    False, '-s', show_default='-!s', help='Silence extreneous messages in console')):
    """Set timestamp in the cache"""
    mtax: MtaxCache = MtaxCache(timezone)
    mtax.get_item('_timestamp_', silent)



from .mtax_cache import MtaxCache
from .cache_app import (
    cache_,
    add_ as add_to_cache,
    incr_ as incr_in_cache,
    set_ as set_item_to_cache,
    get_ as get_item_from_cache,
    decr_ as decr_in_cache,
    clear_ as clear_cache,
    reset_ as reset_cache,
    remove_ as remove_item_from_cache,
    list_ as list_items_in_cache,
    info_ as show_info_of_cache)


__version__ = 'v1.0.0'


__all__ = [
    "MtaxCache",
    "cache_",
    "add_to_cache",
    "incr_in_cache",
    "set_item_to_cache",
    "get_item_from_cache",
    "decr_in_cache",
    "clear_cache",
    "remove_item_from_cache",
    "reset_cache",
    "list_items_in_cache",
    "show_info_of_cache",
    "__version__"]



import concurrent.futures
import time
import sys

from montyalex.us_tools import SETTINGS

TIMEOUT = '1min'
if SETTINGS.get_value('default.opt.timeout'):
    TIMEOUT = SETTINGS.get_value('default.opt.timeout')
MINUTE = False
SECONDS = False

if isinstance(TIMEOUT, str):
    if 'min' in TIMEOUT:
        MINUTE = True
        TIMEOUT = int(TIMEOUT.removesuffix('min')) * 60
if isinstance(TIMEOUT, str):
    if 'sec' in TIMEOUT:
        SECONDS = True
        TIMEOUT = int(TIMEOUT.removesuffix('sec'))

def func_time(message: str = None):
    def decorator(func):
        func_name = f'"{func.__name__}({func.__type_params__})"'
        def wrapper(*args, **kwargs):
            local_message = message if message is not None else 'completed in approximately'

            start_time = time.time()
            with concurrent.futures.ThreadPoolExecutor() as executor:
                future = executor.submit(func, *args, **kwargs)
                try:
                    result = future.result(timeout=TIMEOUT)
                    elapsed_time = time.time() - start_time

                    if local_message and '+' in local_message:
                        func_msg_1, func_msg_2 = [
                            part.strip()
                            for part in local_message.split('+', maxsplit=1)]
                        print(f'{func_name}: {func_msg_1} {elapsed_time:.3f} seconds {func_msg_2}')
                    elif local_message and ';' in local_message:
                        func_msg_1, func_msg_2 = [
                            part.strip()
                            for part in local_message.split(';', maxsplit=1)]
                        print(f'{func_name}: {func_msg_1} {elapsed_time:.3f} seconds {func_msg_2}')
                    else:
                        print(f'{func_name}: {local_message} {elapsed_time:.3f} seconds')

                    return result

                except concurrent.futures.TimeoutError:
                    print(f"Function {func_name} exceeded {TIMEOUT} seconds. Cancelling...")
                    sys.exit("Cancelled due to timeout")
        return wrapper
    return decorator

# ---------------------------------------------------------------------------
# | Example Usage
# ---------------------------------------------------------------------------
if __name__ == '__main__':
    @func_time(message='completed in')
    def short_running_action():
        time.sleep(2)
        print("SHORT Action completed")

    @func_time(message='took + to complete')
    def long_running_action():
        time.sleep(4)
        print("LONG Action completed")

    @func_time(message='took; to complete!')
    def longer_running_action():
        time.sleep(8)
        print("LONGER Action completed")

    @func_time()
    def longest_running_action():
        time.sleep(16)
        print("LONGEST Action completed")

    short_running_action()
    long_running_action()
    longer_running_action()
    longest_running_action()



from ._components import (
    GroupSettingName,
    ShortSettingName,
    SettingName,
    SettingComponent)


# ----------------------------------------------------------------------
# | UserInfo
# ----------------------------------------------------------------------
class UserInfo(SettingComponent):
    def __init__(
        self,
        infix: str,
        suffix: str,
        type_: type,
        *,
        prefix: str = 'user',
        parent: GroupSettingName | ShortSettingName | SettingName = None) -> None:
        super().__init__(prefix, infix, suffix, type_=type_, parent=parent)


# ----------------------------------------------------------------------
# | Locale
# ----------------------------------------------------------------------
locale_continent_userinfo = UserInfo(
    'locale', 'continent', str, parent=SettingName('user', 'locale', 'continent'))
locale_country_userinfo = UserInfo(
    'locale', 'country', str, parent=SettingName('user', 'locale', 'country'))
locale_region_userinfo = UserInfo(
    'locale', 'region', str, parent=SettingName('user', 'locale', 'region'))
locale_state_userinfo = UserInfo(
    'locale', 'state', str, parent=SettingName('user', 'locale', 'state'))
locale_province_userinfo = UserInfo(
    'locale', 'province', str, parent=SettingName('user', 'locale', 'province'))
locale_city_userinfo = UserInfo('locale', 'city', str, parent=SettingName('user', 'locale', 'city'))
locale_streetname_userinfo = UserInfo(
    'locale', 'street-name', str, parent=SettingName('user', 'locale', 'street-name'))
locale_streetnum_userinfo = UserInfo(
    'locale', 'street-number', str, parent=SettingName('user', 'locale', 'street-number'))



from .enumeration import Key, Value
from .exceptions import MontyKeyValueError, MontyJSONError, MontyRangeError
from .pickling import pdumps, ploads
from .singleton import singleton_decorator as singleton


__version__ = 'v1.0.0'


__all__ = [
    "Key",
    "Value",
    "MontyKeyValueError",
    "MontyJSONError",
    "MontyRangeError",
    "pdumps",
    "ploads",
    "singleton",
    "__version__"]



import platform
from diskcache import Cache
import psutil

from montyalex.console_tools import richconsole, success_stm
from montyalex.fs_tools import current_working_dir, joinpaths, pathexists
from montyalex.time_tools import MtaxTime
from montyalex.typing_tools import NoneType
from .cache_helpers import remove_cache__dir

print = richconsole.print


DIR = joinpaths('.mtax', 'cache')
mtax: Cache = Cache(DIR)
os_name = platform.system()
os_strength = platform.processor()
py_ver = platform.python_version()
cache_exists = joinpaths(current_working_dir, DIR, 'cache.db')

svmem = psutil.virtual_memory()

total_ram_bytes = svmem.total
total_ram_gb = total_ram_bytes / (1024 ** 3)

available_ram_bytes = svmem.available
available_ram_gb = available_ram_bytes / (1024 ** 3)

used_ram_bytes = total_ram_bytes - available_ram_bytes
used_ram_gb = used_ram_bytes / (1024 ** 3)

class MtaxCache:
    def __init__(self, timezone: str = 'Etc/Greenwich') -> None:
        self.cache = mtax
        self.status: str = 'Ok'
        self.os_ = os_name
        self.os_strength = os_strength
        self.py_ver = py_ver
        self.total_ram_gb = total_ram_gb
        self.available_ram_gb = available_ram_gb
        self.used_ram_gb = used_ram_gb
        self.datetime: MtaxTime = MtaxTime(timezone)

    def add_item(self, item: tuple[str, str], silent: bool = False):
        if not silent:
            print(f'Adding {item[0]!r} to Cache')
        self.cache.add('_timestamp_', self.datetime.timestamp())
        self.cache.add(item[0], item[1])
        if not silent:
            print(f'{success_stm}, Added {item[0]!r} to Cache')

    def set_item(self, item: tuple[str, str], silent: bool = False):
        if not silent:
            print(f'Setting {item[0]!r} to {item[1]!r} in the Cache')
        self.cache.set('_timestamp_', self.datetime.timestamp())
        self.cache.set(item[0], item[1])
        print(f'{success_stm}, Set {item[0]!r} to {item[1]!r} in the Cache')

    def get_item(self, key_: str, silent: bool = False):
        if not silent:
            print(f'Getting {key_!r} from the Cache')
        item_value = self.cache.get(key_)
        if not silent:
            print(f'{success_stm}, Got {key_!r} from the Cache')
            print(f'The value for {key_!r} in the Cache is {item_value!r}')
        if not silent:
            print(f'{key_!r}={item_value!r}')
        return item_value

    def incr_item(self, key_: str, silent: bool = False):
        if not silent:
            print(f'Incrementing {key_!r} in the Cache')
        item_value = self.cache.incr(key_)
        item_value = max(item_value, 0)
        if not silent:
            print(f'{success_stm}, Incremented {key_!r} in the Cache to {item_value}')
        return item_value

    def decr_item(self, key_: str, silent: bool = False):
        if not silent:
            print(f'Decrementing {key_!r} in the Cache')
        item_value = self.cache.decr(key_)
        item_value = max(item_value, 0)
        if not silent:
            print(f'{success_stm}, Decremented {key_!r} in the Cache to {item_value}')
        return item_value

    def clear(self, silent: bool = False):
        if not silent:
            print('Clearing Cache')
        self.cache.clear()
        remove_cache__dir(DIR, silent)
        if not silent:
            print('Cache Cleared')

    def reset(self, silent: bool = False):
        if not silent:
            print('Resetting Cache')
        self.remove('_python_', silent)
        self.remove('_datedirs_', silent)
        self.remove('_cache-tools_', silent)
        self.remove('_directory-tools_', silent)
        self.add_item(('_python_', f'v{platform.python_version()}'), silent)
        self.add_item(('_datedirs_', 'v1.0.0'), silent)
        self.add_item(('_cache-tools_', 'v1.0.0'), silent)
        self.add_item(('_directory-tools_', 'v1.0.0'), silent)
        if not silent:
            print('Cache Reset')

    def remove(self, key_: str, silent: bool = False):
        if not silent:
            print(f'Removing {key_!r}')
        self.cache.delete(key_)
        if not silent:
            print(f'{success_stm}, Removed {key_!r} from Cache')

    def list_k_v_pairs(self, silent: bool = False):
        cached_items = list(self.cache)
        cached_items.sort()
        for key in cached_items:
            value = self.cache.get(key)
            print(f'[bold]{key}[/]: {value!r}')
        if not silent:
            ...

    def info(self, list_: bool = False, silent: bool = False):
        print("[bold]Info[/]: [")
        print(f"  [bold]Current[/]: {self.datetime.timestamp()!r}")
        if self.status == 'Ok':
            print(f"  [bold]Status[/]: [green]{self.status!r}[/]")
        else:
            print(f"  [bold]Status[/]: [red]{self.status!r}[/]")
        print(f"  [bold]System Processor[/]: {self.os_strength!r}")
        print(f"  [bold]Operating System[/]: {self.os_!r}")
        if self.os_strength and self.os_:
            print(f"    [bold]Python Version[/]: [yellow]'v{self.py_ver}'[/]")
            if self.available_ram_gb:
                print(f"    [bold]RAM Available[/]: [sky_blue2]{self.available_ram_gb:.2f}GB[/]")
            else:
                print(f"    [bold]RAM Available[/]: [orange4]{0:.2f}GB [bold]???[/][/] NONE FOUND")
            if self.used_ram_gb:
                print(f"    [bold]RAM Used[/]: [deep_pink3]{self.used_ram_gb:.2f}GB[/]")
            else:
                print(f"    [bold]RAM Used[/]: [green1]{0:.2f}GB [bold]???[/][/] NONE FOUND")
            if self.total_ram_gb > 0:
                print(f"    [bold]RAM Total[/]: [sky_blue2]{self.total_ram_gb:.2f}GB[/]")
            else:
                print(f"    [bold]RAM Total[/]: [orange4]{0:.2f}GB [bold]???[/][/] NONE FOUND")
            print("    [bold]Status[/]: [green]'Good'[/]")
        else:
            print("    [bold]Status[/]: [red]'Bad'[/]")
        if pathexists(cache_exists):
            print(f'  [bold]Cache Directory[/]: [green]{DIR!r}[/]')
            print("    [bold]File[/]: [green]'cache.db'[/]")
        else:
            print(f'  [bold]Cache Directory[/]: [red]{DIR!r}[/]')
            print("    [bold]File[/]: [red]'cache.db'[/] [dim](does not exist)[/]")
        py_ver_cached = self.get_item('_python_', True)
        cat_ver_cached = self.get_item('_cache-tools_', True)
        cot_ver_cached = self.get_item('_console-tools_', True)
        dir_ver_cached = self.get_item('_directory-tools_', True)
        if isinstance(py_ver_cached, NoneType):
            print(f"    [bold]Python Version[/]: {py_ver_cached!r}")
        else:
            print(f"    [bold]Python Version[/]: [yellow]{py_ver_cached!r}[/]")
        if isinstance(cat_ver_cached, NoneType):
            print(f"    [bold]CacheTools Version[/]: {cat_ver_cached!r}")
        else:
            print(f"    [bold]CacheTools Version[/]: [cyan2]{cat_ver_cached!r}[/]")
        if isinstance(cot_ver_cached, NoneType):
            print(f"    [bold]ConsoleTools Version[/]: {cot_ver_cached!r}")
        else:
            print(f"    [bold]ConsoleTools Version[/]: [cyan2]{cot_ver_cached!r}[/]")
        if isinstance(dir_ver_cached, NoneType):
            print(f"    [bold]DirectoryTools Version[/]: {dir_ver_cached!r}")
        else:
            print(f"    [bold]DirectoryTools Version[/]: [cyan2]{dir_ver_cached!r}[/]")
        if list_:
            print(']..,')
            self.list_k_v_pairs()
        else:
            print(']')
        if not silent:
            ...



from datetime import datetime
from zoneinfo import ZoneInfo, ZoneInfoNotFoundError


class MtaxTime:
    def __init__(self, timezone: str = 'Etc/Greenwich') -> None:
        self.timezone: str = timezone
        try:
            self.tzinfo: ZoneInfo = ZoneInfo(self.timezone)
        except ZoneInfoNotFoundError as exc:
            raise ValueError(f"Invalid timezone name: {self.timezone}") from exc
        self.now: datetime = datetime.now(self.tzinfo)

        self.year: int = self.now.year
        self.month: int = self.now.month
        self.day: int = self.now.day
        self.name: str = self.now.strftime('%a').upper()
        self.daynum_ofweek: int = self.now.weekday()
        self.daynum_ofyear: int = self.now.timetuple().tm_yday

        self.hour: str = f'{self.now.hour:02}'
        self.minute: str = f'{self.now.minute:02}'
        self.second: str = f'{self.now.second:02}'
        self.microsecond: int = self.now.microsecond
        self.millisecond: int = self.microsecond // 1000

        self.tzname: str = self.now.tzname()
        self.tzoffset: str = self.now.strftime('%z')

        self.date: str = f'{self.daynum_ofyear}/{self.year}'
        self.time: str = f'{self.hour}:{self.minute}:{self.second}.{self.microsecond}'

    def __repr__(self) -> str:
        r = (
            f'{self.year}, {self.month}, {self.day}, '
            f'{self.time}, {self.tzname}{self.tzoffset}')
        return f'montyalex.MtaxTime({r})'

    def __str__(self) -> str:
        return self.timestamp()

    def __format__(self, format_spec: str) -> str:
        return self.timestamp(format_spec)

    def timestamp(self, format = None):
        if format is None:
            return f'{self.tzname}/{self.name}-{self.date}T{self.time}{self.tzoffset}'
        return datetime.strftime(self.now, format)



from ._components import (
    GroupSettingName as Group)
from .subcategory import (
    SettingSubcategory as Subcategory,
    option_action_subcategory,
    info_default_subcategory,
    option_default_subcategory,
    locale_userinfo_subcategory)


# ----------------------------------------------------------------------
# | SettingCategory
# ----------------------------------------------------------------------
class SettingCategory:
    def __init__(self, prefix: str, type_: type, *subcategories: Subcategory) -> None:
        self.info: Group = Group(prefix)
        self.type_: type = type_
        self.subcategories: tuple[Subcategory] = subcategories

    def __str__(self) -> str:
        return f'{self.subcategories}'

    def __repr__(self) -> str:
        subcategories = f'{list(self.subcategories)!r}'
        return f'Category(={self.info.prefix}, type={self.type_}, subcategories={subcategories})'

    def add_subcategory(self, subcategory: Subcategory) -> None:
        new_subcategories = (*self.subcategories, subcategory)
        self.subcategories = new_subcategories


# ----------------------------------------------------------------------
# | Categories
# ----------------------------------------------------------------------
ACTION = SettingCategory('action', object, option_action_subcategory)
DEFAULT = SettingCategory('default', object, info_default_subcategory, option_default_subcategory)
USERINFO = SettingCategory('user', object, locale_userinfo_subcategory)



from .json_schema_template import JsonSchemaOptions as schema
from .json_template import JsonOptions as json
from .mpck_template import MpckOptions as mpck
from .toml_template import TomlOptions as toml
from .yaml_template import YamlOptions as yaml


__version__ = 'v1.0.0'


__all__ = [
    "schema",
    "json",
    "mpck",
    "toml",
    "yaml",
    "__version__"]



from .datedirs import create_date_directories as datedirs, remove_date_directories as rmdatedirs
from .simpledirs import (
    create_simple_directories as simpledirs,
    remove_simple_directories as rmsimpledirs)


__version__ = 'v1.0.0'


__all__ = [
    "datedirs",
    "rmdatedirs",
    "simpledirs",
    "rmsimpledirs",
    "__version__"]



from ._mtax import mtax_blueprints, mtax_commands, mtax_complete, mtax_funcargs
from ._settings import settings_
from .cache_tools import cache_, incr_in_cache, decr_in_cache
from .fs_tools import cancel
from .console_tools import richconsole, warn_stm
from .directory_tools import (
    datedirs, rmdatedirs, simpledirs, rmsimpledirs)
from .time_tools import func_time
from .typing_tools import NoneType
from .typer_tools import Option, Typer

print = richconsole.print



monty: Typer = Typer(
    name='monty',
    add_help_option=False,
    pretty_exceptions_show_locals=False)
monty.add_typer(settings_)
monty.add_typer(cache_)
LINEDASH = '___________________________________________________________________________________'

def date_dirs__helper(name_: str, prefix: str, suffix: str, silent: bool = False):
    if (name_ or prefix or suffix) and not silent:
        print('Not allowed with the -datedirs option')
        cancel()

@monty.command(name='mk')
@func_time('to create the directories')
def create__dirs(
    range_: int = Option(1, '--range', '-r'),
    s_dirs: bool = Option(
        False, '--simple-directories', '-dirs', is_flag=True, show_default='range: 1'),
    dt_dirs: bool = Option(
        False, '--date-directories', '-datedirs', is_flag=True, show_default='range: 1yr'),
    four_weeks: bool = Option(False, '--four-weeks', '-4w', is_flag=True),
    dir_parent: str = Option(None, '--parent', '-p'),
    dir_name: str = Option(None, '--name'),
    prefix: str = Option(None, '--prefix'),
    suffix: str = Option(None, '--suffix'),
    silent: bool = Option(
        False, '-s', show_default='-!s', help='Silence extreneous messages in console')):
    """Creates new folders in the current local directory"""
    if s_dirs:
        simpledirs(range_, dir_name, prefix, suffix, silent)
        incr_in_cache(key='_mkdirs', silent=silent)
    if dt_dirs:
        date_dirs__helper(dir_name, prefix, suffix, silent)
        datedirs(range_, four_weeks, dir_parent, silent)
        incr_in_cache(key='_mkdates', silent=silent)
    if not s_dirs and not dt_dirs:
        print('Void, Please provide a type of directory to make')

@monty.command(name='rm')
@func_time('to remove the directories')
def remove__dirs(
    range_: int = Option(1, '--range', '-r'),
    s_dirs: bool = Option(
        False, '--simple-directories', '-dirs', is_flag=True, show_default='range: 1'),
    dt_dirs: bool = Option(
        False, '--date-directories', '-datedirs', is_flag=True, show_default='range: 1yr'),
    four_weeks: bool = Option(False, '--four-weeks', '-4w', is_flag=True),
    dir_parent: str = Option(None, '--parent', '-p'),
    dir_name: str = Option(None, '--name', show_default='000'),
    prefix: str = Option(None, '--prefix'),
    suffix: str = Option(None, '--suffix'),
    silent: bool = Option(
        False, '-s', show_default='-!s', help='Silence extreneous messages in console')):
    """Removes folders in the current local directory"""
    if s_dirs:
        rmsimpledirs(range_, dir_name, prefix, suffix, silent)
        decr_in_cache(key='_mkdirs', silent=silent)
    if dt_dirs:
        date_dirs__helper(dir_name, prefix, suffix, silent)
        rmdatedirs(range_, four_weeks, dir_parent, silent)
        decr_in_cache(key='_mkdates', silent=silent)
    else:
        print('Void, Please provide a type of directory to remove')

def __commands_help(split: bool = False, commands: bool = False):
    if commands:
        if split:
            print('')
        print('cmds.')
        if split:
            print('')
        for name, shorthelp, index, defaults in mtax_commands:
            if isinstance(index, int):
                if index <= 9:
                    command_help = f"([pink3 dim]{index}[/]) {name} -> [dim italic]{shorthelp}[/]"
                else:
                    command_help = f"([pink3 dim]{index}[/]){name} -> [dim italic]{shorthelp}[/]"
            if isinstance(index, NoneType):
                command_help = f"([dim]X[/]) {name} -> [dim italic]{shorthelp}[/]"
            if defaults:
                command_help += f' [pink3 dim italic]({defaults})[/]'
            print(command_help)
            if split:
                print('')

def __funcargs_help(split: bool = False, func_args: bool = False):
    if func_args:
        if split:
            print('')
        print('args.')
        if split:
            print('')
        for func, module, funcargs_index, funcargs_defaults in mtax_funcargs:
            if funcargs_index:
                if funcargs_index != '.':
                    funcargs_help = f"([green1 dim]{funcargs_index}[/]) [italic]{func}[/]"
                else:
                    funcargs_help = f"([dim]{funcargs_index}[/]) [italic]{func}[/]"
            else:
                funcargs_help = f"    [italic]{func}[/]"
            if module:
                funcargs_help += f' -> [green1 dim italic]{module}[/]'
            if funcargs_defaults:
                funcargs_help += f' [dim italic]({funcargs_defaults})[/]'
            if not module and not funcargs_defaults:
                funcargs_help += ' [dim italic]...[/]'
            print(funcargs_help)
            if split:
                print('')

def __blueprints_help(split: bool = False, blueprints: bool = False):
    if blueprints:
        if split:
            print('')
        print('bpns.')
        if split:
            print('')
        for blueprint, module, blueprint_index, blueprint_defaults in mtax_blueprints:
            if blueprint_index:
                blueprint_help = f"([blue3 dim]{blueprint_index}[/]) {blueprint}"
            else:
                blueprint_help = f"    {blueprint}"
            if module:
                blueprint_help += f' -> [blue3 dim italic]{module}[/]'
            if blueprint_defaults:
                blueprint_help += f' [dim italic]({blueprint_defaults})[/]'
            if not module and not blueprint_defaults:
                blueprint_help += ' [dim italic]...[/]'
            print(blueprint_help)
            if split:
                print('')

@monty.command()
@monty.command(name='help')
@cache_.command(name='help')
@settings_.command(name='help')
def mtax(
    help_: bool = Option(False, '--help', '-h'),
    split: bool = Option(False, '--split'),
    save_funcargs_to_file: bool = Option(False, '--save-a', '-s-a'),
    save_blueprints_to_file: bool = Option(False, '--save-b', '-s-b'),
    save_commands_to_file: bool = Option(False, '--save-c', '-s-c'),
    save_file: bool = Option(False, '--save'),
    func_args: bool = Option(False, '-a', '--args'),
    blueprints: bool = Option(False, '-b', '--blueprints'),
    commands: bool = Option(False, '-c', '--commands'),
    super_silence: bool = Option(False, '-ss'),
    silence: bool = Option(False, '-s')):
    """."""
    if help_:
        print('HELP')
    else:
        if not super_silence:
            print(LINEDASH)
            __commands_help(split, (not blueprints and not func_args) or commands)
            if (func_args or blueprints) and commands:
                print(LINEDASH)
            __funcargs_help(split, func_args)
            if func_args and blueprints:
                print(LINEDASH)
            __blueprints_help(split, blueprints)
            print(LINEDASH)
        if save_commands_to_file:
            mtax_complete(commands=True, func_args=False, blueprints=False)
        if save_funcargs_to_file:
            mtax_complete(commands=False, func_args=True, blueprints=False)
        if save_blueprints_to_file:
            mtax_complete(commands=False, func_args=False, blueprints=True)
        if save_file:
            mtax_complete(commands=commands, func_args=func_args, blueprints=blueprints)
            if not silence and not super_silence:
                print(f'{warn_stm}, Syntax highlighting is not supported in files')



from ._components import (
    GroupSettingName,
    ShortSettingName,
    SettingName,
    SettingComponent)
from ._actions import datetime_directory_action, simple_directory_action


# ----------------------------------------------------------------------
# | Default
# ----------------------------------------------------------------------
class Default(SettingComponent):
    def __init__(
        self,
        infix: str,
        suffix: str,
        type_: type,
        *,
        prefix: str = 'default',
        parent: GroupSettingName | ShortSettingName | SettingName = None) -> None:
        super().__init__(prefix, infix, suffix, type_=type_, parent=parent)


# ----------------------------------------------------------------------
# | Defaults
# ----------------------------------------------------------------------
timezone_info_default = Default(
    'info', 'timezone', str, parent= SettingName('default', 'info', 'timezone'))
timeout_option_default = Default('opt', 'timeout', str, parent=SettingName('default', 'opt', 'timeout'))
fourweek_option_default = Default('opt', 'four-week-month', bool, parent=datetime_directory_action)
datetime_range_option_default = Default('opt', 'range', str, parent=datetime_directory_action)
datetime_format_option_default = Default('opt', 'format', str, parent=datetime_directory_action)
simple_range_option_default = Default('opt', 'range', str, parent=simple_directory_action)



from montyalex.console_tools import richconsole
from montyalex.typer_tools import Option, Typer
from .presets import LPR, GPR, MCR, Table
from .treeid import FileSystemTree

print = richconsole.print


tree_: Typer = Typer(
    name='tree',
    pretty_exceptions_show_locals=False)
nodes_: Typer = Typer(
    name='nodes',
    pretty_exceptions_show_locals=False)
tree_.add_typer(nodes_)

@tree_.command(name='create')
def create_(id_: str):
    fsTree = FileSystemTree(id_)
    if fsTree.exists:
        print(f'Tree {id_!r} already exists')
    else:
        fsTree.write()
        print(f'Tree created at: {id_!r}, use setup to continue')

@tree_.command(name='setup')
def setup_(id_: str, template: str = None):
    fsTree = FileSystemTree(id_)
    scannedTree = FileSystemTree(id_).scan()
    if template:
        fsTree.setup(template=template)
    fsTreeName = scannedTree["name"]
    # fsTreeId = scannedTree[""]
    fsTreeAge = scannedTree["age"]
    fsTreeTemplate = None
    if 'template' in scannedTree.keys():
        fsTreeTemplate = scannedTree["template"]
    print(f'Tree: {fsTreeName!r}, {fsTreeAge!r}')
    # print(f'Identifier: {fsTreeId!r}')
    if fsTreeTemplate:
        print(f'Template: {fsTreeTemplate!r}')

@nodes_.command(name='first')
def first_(id_: str):
    fsTree = FileSystemTree(id_).scan()
    fsNodes = fsTree["nodes"]
    firstNode = fsNodes[0]
    print(f'First item of {fsTree["name"]!r} tree: {firstNode}')

@nodes_.command(name='next')
def next_(id_: str):
    fsTree = FileSystemTree(id_).scan()
    print(f'Next item of {fsTree["name"]!r} tree: ')

@nodes_.command(name='prev')
def prev_(id_: str):
    fsTree = FileSystemTree(id_).scan()
    print(f'Prev item of {fsTree["name"]!r} tree: ')

@nodes_.command(name='last')
def last_(id_: str):
    fsTree = FileSystemTree(id_).scan()
    fsNodes = fsTree["nodes"]
    lastNode = fsNodes[-1]
    print(f'Last item of {fsTree["name"]!r} tree: {lastNode}')

@tree_.command(name='table')
def table_(
    least: bool = Option(False, '-LPR', '--least-range'),
    most_common: bool = Option(False, '-MCR', '--common-range'),
    greatest: bool = Option(False, '-GPR', '--greatest-range'),
    show_wz_range: bool = Option(False, '-Z', '--show-zero-range')):
    if least:
        tables: dict[str, Table] = {k: v.table(show_wz_range) for k, v in LPR.items()}
        for table in tables.values():
            table.title = 'L.P. ' + table.title
            print(table)
    if greatest:
        tables: dict[str, Table] = {k: v.table(show_wz_range) for k, v in GPR.items()}
        for table in tables.values():
            table.title = 'G.P. ' + table.title
            print(table)
    if most_common:
        tables: dict[str, Table] = {k: v.table(show_wz_range) for k, v in MCR.items()}
        for table in tables.values():
            table.title = 'M.C. ' + table.title
            print(table)
    if show_wz_range:
        print('The zeroith directory for the digit amount specified is included')
    else:
        print('The zeroith directory for the digit amount specified is reserved')

@tree_.command(name='delete')
def delete_(id_: str):
    fsTree = FileSystemTree(id_)
    if fsTree.exists == True:
        print(f'Tree {id_!r} deleted')
        fsTree.delete()
    else:
        print(f'Tree {id_!r} already does not exist')



from ._components import (
    ShortSettingName as Short,
    SettingComponent as Component)
from ._actions import (
    silent_option_action)
from ._defaults import (
    datetime_range_option_default,
    datetime_format_option_default,
    timeout_option_default,
    fourweek_option_default,
    simple_range_option_default,
    timezone_info_default)
from ._userinfo import (
    locale_city_userinfo,
    locale_continent_userinfo,
    locale_country_userinfo,
    locale_province_userinfo,
    locale_region_userinfo,
    locale_state_userinfo,
    locale_streetname_userinfo,
    locale_streetnum_userinfo)


# ----------------------------------------------------------------------
# | SettingSubcategory
# ----------------------------------------------------------------------
class SettingSubcategory:
    def __init__(self, prefix: str, infix: str, type_: type, *components: Component) -> None:
        self.info: Short = Short(prefix, infix)
        self.type_: type = type_
        self.components: tuple[Component] = components

    def __str__(self) -> str:
        return f'{self.components}'

    def __repr__(self) -> str:
        return (
            f'Subcategory(={self.info.prefix}.{self.info.infix}, '
            f'type={self.type_}, components={list(self.components)!r})'
        )

    def add_component(self, component: Component) -> None:
        new_components = (*self.components, component)
        self.components = new_components


# ----------------------------------------------------------------------
# | Subcategories
# ----------------------------------------------------------------------
option_action_subcategory = SettingSubcategory('action', 'opt', object, silent_option_action)
info_default_subcategory = SettingSubcategory('default', 'info', object, timezone_info_default)
option_default_subcategory = SettingSubcategory('default', 'opt', object,
    timeout_option_default,
    fourweek_option_default,
    datetime_range_option_default,
    datetime_format_option_default,
    simple_range_option_default)
locale_userinfo_subcategory = SettingSubcategory('user', 'locale', object,
    locale_region_userinfo,
    locale_continent_userinfo,
    locale_country_userinfo,
    locale_state_userinfo,
    locale_province_userinfo,
    locale_city_userinfo,
    locale_streetname_userinfo,
    locale_streetnum_userinfo)



import os
import psutil

def virtualmem() -> dict[str, float]:
    total = psutil.virtual_memory().total / (1024 ** 3)
    available = psutil.virtual_memory().available / (1024 ** 3)
    used = psutil.virtual_memory().used / (1024 ** 3)
    recommended_maximum = total / 4

    return {
        'total': total,
        'recommended_maximum': recommended_maximum,
        'available': available,
        'used': used,
    }

class VirtualMemory:
    def __init__(self) -> None:
        self.system = virtualmem()
        self.pid = os.getpid()
        self.process = psutil.Process(self.pid)
        self.memory_info = self.process.memory_info()
        self.used = self.system["used"]
        self.rss = self.memory_info.rss / (1024 ** 2) - self.used

    def __float__(self) -> float:
        return self.system["recommended_maximum"]

    def __call__(self) -> 'VirtualMemory':
        self.system = virtualmem()
        self.memory_info = self.process.memory_info()
        self.rss = self.memory_info.rss / (1024 ** 2) - self.used
        return self

class VirMemProcess:
    def __init__(self) -> None:
        self.available: bool = False
        self.double: bool = False
        self.maximum: bool = False
        self.usage: bool = False
        self.virmem = VirtualMemory()
        self.system = self.virmem.system

    def __call__(
        self,
        *,
        available: bool = False,
        double: bool = False,
        maximum: bool = False,
        usage: bool = False) -> None:
        self.available: bool = available
        self.double: bool = double
        self.maximum: bool = maximum
        self.usage: bool = usage
        self.virmem = VirtualMemory()
        return self

    def __enter__(self):
        self.virmem()

        if self.available:
            self.system["recommended_maximum"] = self.system["available"]
        if self.double:
            self.system["recommended_maximum"] *= 2
        if self.maximum:
            self.system["recommended_maximum"] = self.system["total"]
        if self.usage:
            print(f'System Used: {self.system["used"]:.2f}gb')
            print(f'System Available: {self.system["available"]:.2f}gb')
            print(f'System Total: {self.system["total"]:.2f}gb')
            print(f'Recommended Maximum: {self.system["recommended_maximum"]:.2f}gb')
            print(f'Process Used: {self.virmem.rss / 2:.2f}gb')
        if float(self.virmem) > self.virmem.rss / 2:
            return self
        raise MemoryError('Context manager not allowed due to insufficient available memory')

    def __exit__(self, type, value, traceback):
        pass

vmprocess = VirMemProcess()
__version__ = 'v1.0.0'


__all__ = [
    "vmprocess",
    "__version__"]



from montyalex.time_tools import MtaxTime
from montyalex.typer_tools import Option, Typer
from .mtax_cache import MtaxCache


zone_: Typer = Typer(
    name='zone',
    help='Manage timezone-data in the cache')

@zone_.command(name='reset')
def reset_(
    *,
    timezone: str = 'Etc/Greenwich',
    silent: bool = Option(
    False, '-s', show_default='-!s', help='Silence extreneous messages in console')):
    """Reset timezone in the cache"""
    mtax: MtaxCache = MtaxCache(timezone)
    mtax_time = MtaxTime(timezone)
    mtax.remove('_timezone_', silent)
    set_(timezone=mtax_time.timezone, silent=silent)

@zone_.command(name='set')
def set_(
    *,
    timezone: str = 'Etc/Greenwich',
    silent: bool = Option(
    False, '-s', show_default='-!s', help='Silence extreneous messages in console')):
    """Set timezone in the cache"""
    mtax: MtaxCache = MtaxCache(timezone)
    mtax_time = MtaxTime(timezone)
    mtax.set_item(('_timezone_', mtax_time.timezone), silent)

@zone_.command(name='get')
def get_(
    *,
    silent: bool = Option(
    False, '-s', show_default='-!s', help='Silence extreneous messages in console')):
    """Set timezone in the cache"""
    mtax: MtaxCache = MtaxCache('Etc/Greenwich')
    mtax.get_item('_timezone_', silent)



from ._components import (
    GroupSettingName,
    ShortSettingName,
    SettingName,
    SettingComponent)


# ----------------------------------------------------------------------
# | Action
# ----------------------------------------------------------------------
class Action(SettingComponent):
    def __init__(
        self,
        infix: str,
        suffix: str,
        type_: type,
        *,
        prefix: str = 'action',
        parent: GroupSettingName | ShortSettingName | SettingName = None) -> None:
        super().__init__(prefix, infix, suffix, type_=type_, parent=parent)


# ----------------------------------------------------------------------
# | Actions
# ----------------------------------------------------------------------
datetime_directory_action = Action(
    'dirs', 'datetime', object, parent=SettingName('action', 'dirs', 'datetime'))
simple_directory_action = Action(
    'dirs', 'simple', object, parent=SettingName('action', 'dirs', 'simple'))
silent_option_action = Action('opt', 'silent', bool, parent=SettingName('action', 'opt', 'silent'))



from datetime import datetime, timedelta
from dateutil.relativedelta import relativedelta

from montyalex.cache_tools import add_to_cache, remove_item_from_cache
from montyalex.console_tools import richconsole, success_stm
from montyalex.fs_tools import abspath, current_working_dir, joinpaths, mkdirs, pathexists
from montyalex.us_tools import datetime_format_option_default
from .helpdirs import remove_formated__dir

print = richconsole.print


FORMAT = datetime_format_option_default.get_value()
if FORMAT == 'Notfound':
    FORMAT = '%Y-%m-%d'
START = datetime.now()

def create_date_directories(
        range_: int = 1, four_weeks: bool = False, parent: str = None, silent: bool = False):
    add_to_cache(key='_datedirs_', value=f'{START:FORMAT}', silent=True)
    new_date = START
    for i in range(range_ * 12):
        if four_weeks:
            new_date += timedelta(days=28)
        else:
            new_date = START + relativedelta(months=i)
        formatted_date = new_date.strftime(FORMAT)
        if parent:
            mkdirs(joinpaths(current_working_dir, parent, formatted_date), exist_ok=True)
        else:
            mkdirs(joinpaths(current_working_dir, formatted_date), exist_ok=True)
        if not silent:
            cwd_ = abspath(parent) if parent else current_working_dir
            print(
                f'{success_stm}, Created {formatted_date!r} in {cwd_!r}')

def remove_date_directories(
        range_: int = 1, four_weeks: bool = False, parent: str = None, silent: bool = False):
    remove_item_from_cache(key='_datedirs_', silent=True)
    if parent:
        remove_formated__dir(parent, silent)

        new_date = START
        for i in range(range_ * 12):
            if four_weeks:
                new_date += timedelta(days=28)
            else:
                new_date = START + relativedelta(months=i)
            formatted_date = new_date.strftime('%Y-%m-%d')
            formatted_directory = joinpaths(current_working_dir, formatted_date)
            formatted_directory = joinpaths(current_working_dir, parent)
            if not pathexists(formatted_directory) and not silent:
                print(
                    f'{success_stm}, Removed {formatted_date!r} from {formatted_directory!r}')
        if not pathexists(parent) and not silent:
            print(f'{success_stm}, Removed {parent!r} from {current_working_dir!r}')
    else:
        for i in range(range_ * 12):
            new_date = START + relativedelta(months=i)
            formatted_date = new_date.strftime('%Y-%m-%d')
            formatted_directory = joinpaths(current_working_dir, formatted_date)
            remove_formated__dir(formatted_directory, silent)
            if not pathexists(formatted_directory) and not silent:
                print(f'{success_stm}, Removed {formatted_date!r} from {current_working_dir!r}')



from montyalex.uo_tools._useroptions import UserOptions


class TomlOptions(UserOptions):
    def __init__(self, *, directory: str = '.mtax', filename: str = 'settings') -> None:
        super().__init__(
            json=False, toml=True, yaml=False, mpck=False, directory=directory, filename=filename)



from json import JSONDecodeError
from montyalex.typing_tools import Any


class MontyBaseException(BaseException):
    def __init__(self, message: str, value: Any = None, name: str = 'MontyBaseException') -> None:
        super().__init__(message)
        self.name = name
        self.message = message
        self.value = value

    def __str__(self) -> str:
        return f'{self.name}: {self.message} {self.value}'

    def __repr__(self) -> str:
        return self.name

    def __reduce__(self) -> tuple[Any, tuple[str, Any]]:
        return self.__class__, (self.message, self.value)

class MontyArtithmeticException(ArithmeticError):
    def __init__(
        self,
        message: str,
        value: Any = None,
        name: str = 'MontyArtithmeticException') -> None:
        super().__init__(message)
        self.name = name
        self.message = message
        self.value = value

    def __str__(self) -> str:
        return f'{self.name}: {self.message} {self.value}'

    def __repr__(self) -> str:
        return self.name

    def __reduce__(self) -> tuple[Any, tuple[str, Any]]:
        return self.__class__, (self.message, self.value)

class MontyRangeError(MontyArtithmeticException):
    def __init__(self, message: str, value: Any = None, name: str = 'RangeError') -> None:
        super().__init__(message)
        self.name = name
        self.message = message
        self.value = value

class MontyKeyValueError(MontyBaseException): ...
class MontyJSONError(JSONDecodeError): ...



from .console_tools import richconsole
from .typer_tools import Option, Typer
from .uo_tools import schema, json, toml, yaml, mpck
from .us_tools import SETTINGS

print = richconsole.print


settings_: Typer = Typer(
    name='settings',
    add_help_option=False,
    pretty_exceptions_show_locals=False)

init_obj = {
    "action.dirs.datetime": {
        "default.opt.four-week-month": False,
        "default.opt.format": "%Y-%m-%d",
        "default.opt.range": "1yr"
    },
    "action.dirs.simple": {
        "default.opt.range": 1
    },
    "action.opt.silent": False,
    "default.info.timezone": None,
    "user.locale.city": None,
    "user.locale.continent": None,
    "user.locale.country": None,
    "user.locale.state": None,
    "user.locale.street-name": None,
    "user.locale.street-number": None
}

schema_obj = {
    "description": "",
    "type": "object",
    "properties": {
        "action.dirs.datetime": {
            "description": "",
            "type": "object",
            "properties": {
                "default.opt.four-week-month": {
                    "description": "Use a four week (28d) range instead of a complete month",
                    "type": ["boolean", "null"],
                    "default": False
                },
                "default.opt.format": {
                    "description": "Change the date format used for folders",
                    "type": ["string", "null"],
                    "default": "%Y-%m-%d"
                },
                "default.opt.range": {
                    "description": "",
                    "type": ["string", "null"],
                    "default": "1yr"
                }
            }
        },
        "action.dirs.simple": {
            "description": "",
            "type": "object",
            "properties": {
                "default.opt.range": {
                    "description": "",
                    "type": ["integer", "null"],
                    "default": 1
                }
            }
        },
        "action.opt.silent": {
            "description": "",
            "type": ["boolean", "null"],
            "default": False
        },
        "default.info.timezone": {
            "description": "Default timezone used for date related commands",
            "type": ["string", "null"]
        }
    }
}

def _schema_init(dirname: str = '.mtax', filename: str = 'schema'):
    _schema = schema(directory=dirname, filename=filename)
    _schema.change('$schema', 'http://json-schema.org/draft-07/schema#')
    _schema.change('$id', 'montyalex.python-cli.settings-v0.1.2')
    _schema.change('title', 'Settings')
    _schema.change('description', 'MontyAlex CLI Settings')
    _schema.change('type', 'object')
    _schema.change('properties', {"mtax": schema_obj})
    _schema.change('required', ["mtax"])

@settings_.command(name='init', add_help_option=False)
def init_(
    dirname: str = Option('.mtax', '--directory-name', '-dir'),
    filename: str = Option('settings', '--file-name', '-name'),
    *,
    overwrite: bool = Option(False, '--overwrite', '-o'),
    use_all: bool = Option(False, '--all', '-a'),
    use_synced: bool = Option(False, '--sync', '-s'),
    use_json: bool = Option(False, '--json', '-j'),
    use_toml: bool = Option(False, '--toml', '-t'),
    use_yaml: bool = Option(False, '--yaml', '-y'),
    use_mpck: bool = Option(False, '--mpck', '-m')):
    uo: json | toml | yaml | mpck | None = None
    if use_json:
        uo = json(directory=dirname, filename=filename)
    if use_toml:
        uo = toml(directory=dirname, filename=filename)
    if use_yaml:
        uo = yaml(directory=dirname, filename=filename)
    if use_mpck:
        uo = mpck(directory=dirname, filename=filename)
    if uo is None:
        uo = json(directory=dirname, filename=filename)

    if not use_all or use_synced:
        if overwrite:
            if isinstance(uo, json):
                uo.change('$schema', './schema.json')
                _schema_init(dirname)
            uo.change('mtax', init_obj, append=False)
        else:
            if isinstance(uo, json):
                uo.change('$schema', './schema.json')
                _schema_init(dirname)
            if isinstance(uo, toml):
                uo.change('mtax', {
                    "action.dirs.datetime": {
                        "default.opt.four-week-month": False,
                        "default.opt.format": "%Y-%m-%d",
                        "default.opt.range": "1yr"
                    },
                    "action.dirs.simple": {
                        "default.opt.range": 1
                    },
                    "action.opt.silent": False,
                    "default.info.timezone": "",
                    "user.locale.city": "",
                    "user.locale.continent": "",
                    "user.locale.country": "",
                    "user.locale.state": "",
                    "user.locale.street-name": "",
                    "user.locale.street-number": ""
                })
            else:
                uo.change('mtax', init_obj)
    else:
        juo = json(directory=dirname, filename=filename)
        tuo = toml(directory=dirname, filename=filename)
        yuo = yaml(directory=dirname, filename=filename)
        muo = mpck(directory=dirname, filename=filename)
        if overwrite:
            juo.change('$schema', './schema.json')
            _schema_init(dirname)
            juo.change('mtax', init_obj)
            tuo.change('mtax', {
                "action.dirs.datetime": {
                    "default.opt.four-week-month": False,
                    "default.opt.range": "1yr"
                },
                "action.dirs.simple": {
                    "default.opt.range": 1
                },
                "action.opt.silent": False,
                "default.info.timezone": "",
                "user.locale.city": "",
                "user.locale.continent": "",
                "user.locale.country": "",
                "user.locale.state": "",
                "user.locale.street-name": "",
                "user.locale.street-number": ""
            }, append=False)
            yuo.change('mtax', init_obj, append=False)
            muo.change('mtax', init_obj, append=False)
        else:
            juo.change('$schema', './schema.json')
            _schema_init(dirname)
            juo.change('mtax', init_obj)
            tuo.change('mtax', {
                "action.dirs.datetime": {
                    "default.opt.four-week-month": False,
                    "default.opt.range": "1yr"
                },
                "action.dirs.simple": {
                    "default.opt.range": 1
                },
                "action.opt.silent": False,
                "default.info.timezone": "",
                "user.locale.city": "",
                "user.locale.continent": "",
                "user.locale.country": "",
                "user.locale.state": "",
                "user.locale.street-name": "",
                "user.locale.street-number": ""
            })
            yuo.change('mtax', init_obj)
            muo.change('mtax', init_obj)

@settings_.command(name='list', add_help_option=False)
def list_(
    repr_listing: bool = Option(False, '--repr', '-r'),
):
    if repr_listing:
        print(f'{SETTINGS!r}')
    else:
        SETTINGS.show_all_values()

@settings_.command(name='delete', add_help_option=False)
def delete_(
    dirname: str = Option('.mtax', '--directory-name', '-dir'),
    filename: str = Option('settings', '--file-name', '-name'),
    key: str = Option(None, '--key', '-k'),
    *,
    use_json: bool = Option(False, '--json', '-j'),
    use_toml: bool = Option(False, '--toml', '-t'),
    use_yaml: bool = Option(False, '--yaml', '-y'),
    use_mpck: bool = Option(False, '--mpck', '-m')):
    uo: json | toml | yaml | mpck | None = None
    if use_json:
        uo = json(directory=dirname, filename=filename)
    if use_toml:
        uo = toml(directory=dirname, filename=filename)
    if use_yaml:
        uo = yaml(directory=dirname, filename=filename)
    if use_mpck:
        uo = mpck(directory=dirname, filename=filename)
    if uo is None:
        uo = json(directory=dirname, filename=filename)

    if key:
        uo.change(key, None, overwrite=True)
    uo.remove()

def inspect_initalize(uo, dirname, filename, use_json, use_toml, use_yaml, use_mpck):
    if not uo.exists:
        print('Creating new settings file...')
        init_(
            dirname=dirname,
            filename=filename,
            overwrite=False,
            use_json=use_json,
            use_toml=use_toml,
            use_yaml=use_yaml,
            use_mpck=use_mpck)

@settings_.command(name='inspect', add_help_option=False)
def inspect_(
    dirname: str = Option('.mtax', '--directory-name', '-dir'),
    filename: str = Option('settings', '--file-name', '-name'),
    use_json: bool = Option(False, '--json', '-j'),
    use_toml: bool = Option(False, '--toml', '-t'),
    use_yaml: bool = Option(False, '--yaml', '-y'),
    use_mpck: bool = Option(False, '--mpck', '-m'),
    full_inspection: bool = Option(False, '--full', '-fi'),
    mem_alloc_inspection: bool = Option(False, '--memory', '-mi'),
    repr_inspection: bool = Option(False, '--repr', '-ri'),
    exists_inspection: bool = Option(False, '--exists', '-ei'),
    key_inspection: str = Option(None, '--key', '-ki'),
):
    uo: json | toml | yaml | mpck | None = None
    if use_json:
        juo: json = json(directory=dirname, filename=filename)
        uo = juo
    if use_toml:
        tuo: toml = toml(directory=dirname, filename=filename)
        uo = tuo
    if use_yaml:
        yuo: yaml = yaml(directory=dirname, filename=filename)
        uo = yuo
    if use_mpck:
        muo: mpck = mpck(directory=dirname, filename=filename)
        uo = muo
    if uo is None:
        uo = json(directory=dirname, filename=filename)

    inspection: bool = False
    inspect_initalize(uo, dirname, filename, use_json, use_toml, use_yaml, use_mpck)
    if full_inspection:
        inspection = True
        uo.inspect(full=True)
        print("[green]Verified Settings:[/]")
        print(f'{SETTINGS!r}')
        SETTINGS.show_all_values()
        # if repr_listing:
        #     print(f'{SETTINGS!r}')
        # else:
        #     print(f'{SETTINGS!r}')
        #     SETTINGS.show_all_values()
    if mem_alloc_inspection:
        inspection = True
        uo.inspect(mem_alloc=True)
    if repr_inspection:
        inspection = True
        uo.inspect(representation=True)
    if key_inspection:
        inspection = True
        uo.inspect(key=key_inspection)
    if exists_inspection:
        inspection = True
        if uo.exists:
            print(f'Found! [green]{uo.modelpath}[/]')
        else:
            print(f'Not Found! [red]{uo.modelpath}[/]')
    if not inspection:
        uo.inspect()



import os
from posixpath import abspath, expanduser
import shutil


joinpaths: callable = os.path.join

def mkdirs(
    name,
    mode: int = 511,
    exist_ok: bool = False
):
    os.makedirs(name=name, mode=mode, exist_ok=exist_ok)

def rmfile(
    path,
):
    os.remove(path=path)

def rmdirs(
    name,
):
    os.removedirs(name=name)

def rmtree(
    path,
):
    shutil.rmtree(path=path)

pathexists: callable = os.path.exists
_: callable = abspath, expanduser

current_working_dir: str = os.getcwd()



from montyalex.console_tools import richconsole
from montyalex.typing_tools import Any
from .category import (
    SettingCategory as Category,
    ACTION,
    DEFAULT,
    USERINFO)

print = richconsole.print


class Settings:
    def __init__(self) -> None:
        self.action: Category = ACTION
        self.default: Category = DEFAULT
        self.userinfo: Category = USERINFO

        self.all: list[Category] = [
            self.action,
            self.default,
            self.userinfo
        ]

    def __str__(self) -> str:
        return f'{self.all}'

    def __repr__(self) -> str:
        return f'Settings({self.all})'

    def get_value(self, key: str, value_only: bool = True) -> Any:
        for category in self.all:
            for subcat in category.subcategories:
                for comp in subcat.components:
                    if f'{comp}' == key:
                        value = comp.get_value()
                        if value_only:
                            return value
                        if f'{comp.parent}' != f'{comp}':
                            return f'mtax.{comp.parent}.{comp}: {value}'
                        return f'mtax.{comp}: {value}'
        return None

    def show_value(self, key: str) -> None:
        for category in self.all:
            for subcat in category.subcategories:
                for comp in subcat.components:
                    if f'{comp}' == key:
                        value = comp.get_value()
                        if f'{comp.parent}' != f'{comp}':
                            print(f'mtax.{comp.parent}.{comp}: {value}')
                        else:
                            print(f'mtax.{comp}: {value}')

    def show_all_values(self) -> None:
        components: list[Category] = []
        for category in self.all:
            for subcat in category.subcategories:
                for component in subcat.components:
                    components.append(component)
        components.sort(
            key=lambda c: (c.parent.prefix.value, c.parent.infix.value, c.parent.suffix.value)
        )
        for comp in components:
            value = comp.get_value()
            type_ = f'{comp.type_}'.removeprefix('<class ').removesuffix('>')
            if f'{comp.parent}' != f'{comp}':
                if value != 'Notfound':
                    print(
                        f'[yellow italic]mtax.[/][orange1]{comp.parent}[/]'
                        f'.[yellow]{comp}[/]: {value!r}'
                    )
                else:
                    print(
                        f'[yellow italic]mtax.[/][orange1]{comp.parent}[/]'
                        f'.[yellow]{comp}[/]: [red]{value}[/]'
                        f'[dim] (Value should be of type {type_})[/]'
                    )
            else:
                if value != 'Notfound':
                    print(f'[yellow italic]mtax.[/][yellow]{comp}[/]: {value!r}')
                else:
                    print(
                        f'[yellow italic]mtax.[/][yellow]{comp}[/]: [red]{value}[/]'
                        f'[dim] (Value should be of type {type_})[/]'
                    )


# ----------------------------------------------------------------------
# | Global Settings Object
# ----------------------------------------------------------------------
SETTINGS = Settings()



from abc import abstractmethod
from functools import lru_cache
import msgspec

from montyalex.fs_tools import current_working_dir, joinpaths, pathexists, mkdirs
from montyalex.typing_tools import Any, Dict, Optional


class UserTemplate:
    def __init__(
        self,
        json: bool = False,
        toml: bool = False,
        yaml: bool = False,
        mpck: bool = False,
        directory: str = '.mtax',
        filename: str = 'settings') -> None:
        self._cwd: str = current_working_dir

        self.dirname: str = directory
        self.filename: str = filename

        self.dirpath: str = joinpaths(self._cwd, self.dirname)
        self.filepath: str = joinpaths(self.dirpath, self.filename)

        self._dir_exists: str = pathexists(self.dirpath)
        self._file_exists: str = pathexists(self.filepath)

        selected_formats = [fmt for fmt, enabled in {
            'json': json,
            'toml': toml,
            'yaml': yaml,
            'mpck': mpck
        }.items() if enabled]

        if not selected_formats:
            raise ValueError("At least one format must be enabled (json, toml, yaml, mpck).")

        if len(selected_formats) > 1:
            raise ValueError("Only one format can be enabled at a time.")

        selected_format = selected_formats[0]

        self.modelname = selected_format
        self.modelpath = f'{self.filepath}.{self.modelname}'
        match self.modelname:
            case 'json':
                self.model = self.json
            case 'toml':
                self.model = self.toml
            case 'yaml':
                self.model = self.yaml
            case 'mpck':
                self.model = self.mpck

    @property
    def exists(self) -> bool:
        return pathexists(self.modelpath)

    @property
    @abstractmethod
    def options(self) -> Dict[str, Any]:
        raise NotImplementedError

    @property
    @abstractmethod
    def unpacked(self) -> Dict[str, Any]:
        raise NotImplementedError

    @property
    @abstractmethod
    def items(self) -> Dict[str, Any]:
        raise NotImplementedError

    @abstractmethod
    def inspect(self):
        raise NotImplementedError

    @lru_cache(maxsize=16)
    def __read_settings(self) -> Dict[str, Any]:
        if pathexists(self.modelpath):
            with open(self.modelpath, 'rb') as settings_file:
                data = settings_file.read()
            if self.modelname == 'json':
                return self._djson(data)
            if self.modelname == 'toml':
                return self._dtoml(data)
            if self.modelname == 'yaml':
                return self._dyaml(data)
            if self.modelname == 'mpck':
                return self._dmpck(data)
        return {}

    def __save_settings(
        self,
        encoded_data: bytes) -> None:
        if not self._dir_exists:
            mkdirs(self.dirpath)
        with open(self.modelpath, 'wb') as settings_file:
            settings_file.write(encoded_data)

    def convert(self, *args, **kwargs):
        msgspec.convert(*args, **kwargs)

    def json(
        self, data: Optional[Any] = None,
        read: bool = False,
        write: bool = False,
        append: bool = False, overwrite: bool = False) -> Dict[str, Any] | None:
        if read:
            return self.__read_settings()
        if write and data is not None:
            self._ejson(data, append=append, overwrite=overwrite)
        return None
    def _ejson(self, data: Dict[str, Any], append: bool = False, overwrite: bool = False) -> None:
        if append:
            existing = self.__read_settings()
            existing.update(data)
            encoded = msgspec.json.encode(existing)
        elif overwrite or not self.exists:
            encoded = msgspec.json.encode(data)
        else:
            encoded = msgspec.json.encode(data)
        encoded = msgspec.json.format(encoded, indent=4)
        self.__save_settings(encoded)
    def _djson(self, *args, **kwargs):
        return msgspec.json.decode(*args, **kwargs)

    def toml(
        self, data: Optional[Any] = None,
        read: bool = False,
        write: bool = False,
        append: bool = False, overwrite: bool = False) -> Dict[str, Any] | None:
        if read:
            return self.__read_settings()
        if write and data is not None:
            self._etoml(data, append=append, overwrite=overwrite)
        return None
    def _etoml(self, data: Dict[str, Any], append: bool = False, overwrite: bool = False) -> None:
        if append:
            existing = self.__read_settings()
            existing.update(data)
            encoded = msgspec.toml.encode(existing)
        elif overwrite or not self.exists:
            encoded = msgspec.toml.encode(data)
        else:
            encoded = msgspec.toml.encode(data)
        self.__save_settings(encoded)
    def _dtoml(self, *args, **kwargs):
        return msgspec.toml.decode(*args, **kwargs)

    def yaml(
        self, data: Optional[Any] = None,
        read: bool = False,
        write: bool = False,
        append: bool = False, overwrite: bool = False) -> Dict[str, Any] | None:
        if read:
            return self.__read_settings()
        if write and data is not None:
            self._eyaml(data, append=append, overwrite=overwrite)
        return None
    def _eyaml(self, data: Dict[str, Any], append: bool = False, overwrite: bool = False) -> None:
        if append:
            existing = self.__read_settings()
            existing.update(data)
            encoded = msgspec.yaml.encode(existing)
        elif overwrite or not self.exists:
            encoded = msgspec.yaml.encode(data)
        else:
            encoded = msgspec.yaml.encode(data)
        self.__save_settings(encoded)
    def _dyaml(self, *args, **kwargs):
        return msgspec.yaml.decode(*args, **kwargs)

    def mpck(
        self, data: Optional[Any] = None,
        read: bool = False,
        write: bool = False,
        append: bool = False, overwrite: bool = False) -> Dict[str, Any] | None:
        if read:
            return self.__read_settings()
        if write and data is not None:
            self._empck(data, append=append, overwrite=overwrite)
        return None
    def _empck(self, data: Dict[str, Any], append: bool = False, overwrite: bool = False) -> None:
        if append:
            existing = self.__read_settings()
            existing.update(data)
            encoded = msgspec.msgpack.encode(existing)
        elif overwrite or not self.exists:
            encoded = msgspec.msgpack.encode(data)
        else:
            encoded = msgspec.msgpack.encode(data)
        self.__save_settings(encoded)
    def _dmpck(self, *args, **kwargs):
        return msgspec.msgpack.decode(*args, **kwargs)


# ---------------------------------------------------------------------------
# | Example Usage
# ---------------------------------------------------------------------------
if __name__ == '__main__':
    standard_user_template = UserTemplate(json=True)
    std_temp = standard_user_template

    d: Dict[str, Any] = {"": ""}
    std_temp.model(d, write=True)



from montyalex.fs_tools import cancel
from montyalex.object_tools import Key, Value, MontyRangeError as RangeError
from montyalex.typing_tools import Any, Literal
from montyalex.uo_tools import yaml


def statemanager_decorator(
        key: Key,
        value: Value,
        file_name: str,
        *,
        loc: str = None):
    yaml_state = yaml(directory=loc if loc else '.mtax', filename=file_name)
    def statemanager(func):
        state: dict[Key, Any] = {key: value}
        def wrapper():
            print(f"{state}")
            func()
        yaml_state.model(state, write=True, overwrite=True)
        return wrapper
    print(yaml_state.modelpath)
    return statemanager

def statemodifier(
    value: Value,
    operation: Literal[0] | Literal[1],
    multiplier: int = 1,
    limit: int | tuple[int, int] = None,
    ilimit: int | tuple[int, int] = None):
    if isinstance(limit, tuple) or isinstance(ilimit, tuple):
        range_limit = range(*limit) if not ilimit else range(ilimit[0], ilimit[1] + 1)
        check = value + multiplier if operation == 0 else value - multiplier
        if value not in range_limit or check not in range_limit:
            if limit:
                raise RangeError(f'Index ({check}) not in', range(*limit))
            if ilimit:
                raise RangeError(f'Index ({check}) not in', range(*ilimit))
            cancel()
    if isinstance(limit, int) or isinstance(ilimit, int):
        range_limit = range(limit) if not ilimit else range(ilimit + 1)
        check = value + multiplier if operation == 0 else value - multiplier
        if value not in range_limit or check not in range_limit:
            if limit:
                raise RangeError(f'Index ({check}) not in', range(limit))
            if ilimit:
                raise RangeError(f'Index ({check}) not in', range(ilimit))
            cancel()
    return True

def statedecr_decorator(
        key: Key,
        file_name: str,
        *,
        multiplier: int = 1,
        limit: int | tuple[int, int] = None,
        ilimit: int | tuple[int, int] = None,
        loc: str = None):
    yaml_state = yaml(directory=loc if loc else '.mtax', filename=file_name)
    def statedecrementer(func):
        key_value = yaml_state.model(read=True)[key]
        state: dict[Key, Any] = {}
        if isinstance(key_value, int):
            state: dict[Key, Any] = {key: key_value - multiplier}
        def wrapper():
            func()
        try:
            smdecr = statemodifier(key_value, 1, multiplier, limit, ilimit)
            if state[key] != key_value and smdecr:
                print(f"{state}")
                yaml_state.model(state, write=True, overwrite=True)
                print(yaml_state.modelpath)
        except RangeError as e:
            print(e)
        return wrapper
    return statedecrementer

def stateincr_decorator(
        key: Key,
        file_name: str,
        *,
        multiplier: int = 1,
        limit: int | tuple[int, int] = None,
        ilimit: int | tuple[int, int] = None,
        loc: str = None):
    yaml_state = yaml(directory=loc if loc else '.mtax', filename=file_name)
    def stateincrementer(func):
        key_value = yaml_state.model(read=True)[key]
        state: dict[Key, Any] = {}
        if isinstance(key_value, int):
            state: dict[Key, Any] = {key: key_value + multiplier}
        def wrapper():
            func()
        try:
            smincr = statemodifier(key_value, 0, multiplier, limit, ilimit)
            if state[key] != key_value and smincr:
                print(f"{state}")
                yaml_state.model(state, write=True, overwrite=True)
                print(yaml_state.modelpath)
        except RangeError as e:
            print(e)
        return wrapper
    return stateincrementer



from abc import abstractmethod
from functools import total_ordering
from json import JSONDecodeError

from montyalex.fs_tools import pathexists, joinpaths, current_working_dir
from montyalex.typing_tools import Any
from montyalex.uo_tools import json, toml, yaml, mpck


class _Prefix:
    def __init__(self, value: str) -> None:
        self.value = value

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return f'Prefix({self.value})'

class _Infix:
    def __init__(self, value: str) -> None:
        self.value = value

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return f'Infix({self.value})'

class _Suffix:
    def __init__(self, value: str) -> None:
        self.value = value

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return f'Suffix({self.value})'

@total_ordering
class GroupSettingName:
    def __init__(self, prefix: str) -> None:
        self.prefix: _Prefix = _Prefix(prefix)

    @abstractmethod
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SettingName):
            return NotImplemented
        return self.prefix.value == other.prefix.value

    @abstractmethod
    def __lt__(self, other: object) -> bool:
        if not isinstance(other, SettingName):
            return NotImplemented
        if self.prefix.value < other.prefix.value:
            return True
        return False

    @abstractmethod
    def __hash__(self):
        return hash((self.prefix.value))

@total_ordering
class ShortSettingName(GroupSettingName):
    def __init__(self, prefix: str, infix: str) -> None:
        super().__init__(prefix)
        self.infix: _Infix = _Infix(infix)

    @abstractmethod
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SettingName):
            return NotImplemented
        return (self.prefix.value, self.infix.value) == (other.prefix.value, other.infix.value)

    @abstractmethod
    def __lt__(self, other: object) -> bool:
        if not isinstance(other, SettingName):
            return NotImplemented
        if self.prefix.value < other.prefix.value:
            return True
        if self.prefix.value > other.prefix.value:
            return False
        return self.infix.value < other.infix.value

    @abstractmethod
    def __hash__(self):
        return hash((self.prefix.value, self.infix.value))

class SettingName(ShortSettingName):
    def __init__(self, prefix: str, infix: str, suffix: str) -> None:
        super().__init__(prefix, infix)
        self.suffix: _Suffix = _Suffix(suffix)

    def __str__(self) -> str:
        return f'{self.prefix}.{self.infix}.{self.suffix}'

    def __repr__(self) -> str:
        return f'Name({self.prefix}.{self.infix}.{self.suffix})'

    @abstractmethod
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SettingName):
            return NotImplemented
        return (self.prefix.value, self.infix.value) == (other.prefix.value, other.infix.value)

    @abstractmethod
    def __lt__(self, other: object) -> bool:
        if not isinstance(other, SettingName):
            return NotImplemented
        if self.prefix.value < other.prefix.value:
            return True
        if self.prefix.value > other.prefix.value:
            return False
        if self.infix.value < other.infix.value:
            return True
        if self.infix.value > other.infix.value:
            return False
        return self.suffix.value < other.suffix.value

    @abstractmethod
    def __hash__(self):
        return hash((self.prefix.value, self.infix.value, self.suffix.value))

class SettingComponent(SettingName):
    def __init__(
        self, prefix: str, infix: str, suffix: str,
        *,
        parent: (GroupSettingName | ShortSettingName | SettingName) = None,
        type_: type) -> None:
        super().__init__(prefix, infix, suffix)
        self.parent: GroupSettingName | ShortSettingName | SettingName = parent
        self.type_: type = type_
        self.value: str | bool | int | None = None

    def __str__(self) -> str:
        return f'{self.prefix}.{self.infix}.{self.suffix}'

    def __repr__(self) -> str:
        if self.get_value() != 'Notfound':
            return f'Component(={self!s}, type={self.type_}, value={self.get_value()!r})'
        return f'Component(={self!s}, type={self.type_}, value={self.get_value()})'

    def __eq__(self, other: object) -> bool:
        parent_name = f"{self.parent.prefix}.{self.parent.infix}.{self.parent.suffix}"
        if parent_name != f"{self!s}":
            if not isinstance(other, SettingName):
                return NotImplemented
            return (self.prefix.value, self.infix.value) == (other.prefix.value, other.infix.value)
        if not isinstance(other, SettingName):
            return NotImplemented
        return (self.parent.prefix.value, self.parent.infix.value) == \
            (other.parent.prefix.value, other.parent.infix.value)

    def __lt__(self, other: object) -> bool:
        parent_name = f"{self.parent.prefix}.{self.parent.infix}.{self.parent.suffix}"
        if parent_name != f"{self!s}":
            if not isinstance(other, SettingName):
                return NotImplemented
            if self.prefix.value < other.prefix.value:
                return True
            if self.prefix.value > other.prefix.value:
                return False
            if self.infix.value < other.infix.value:
                return True
            if self.infix.value > other.infix.value:
                return False
            return self.suffix.value < other.suffix.value
        if not isinstance(other, SettingName):
            return NotImplemented
        if self.parent.prefix.value < other.parent.prefix.value:
            return True
        if self.parent.prefix.value > other.parent.prefix.value:
            return False
        if self.parent.infix.value < other.parent.infix.value:
            return True
        if self.parent.infix.value > other.parent.infix.value:
            return False
        return self.parent.suffix.value < other.parent.suffix.value

    def __hash__(self):
        return hash((self.prefix.value, self.infix.value, self.suffix.value))

    def get_value(self) -> Any:
        uo: json | toml | yaml | mpck | None = None
        uo_exists = False
        if pathexists(joinpaths(current_working_dir, '.mtax', 'settings.json')):
            uo = json()
        if pathexists(joinpaths(current_working_dir, '.mtax', 'settings.toml')):
            uo = toml()
        if pathexists(joinpaths(current_working_dir, '.mtax', 'settings.yaml')):
            uo = yaml()
        if pathexists(joinpaths(current_working_dir, '.mtax', 'settings.mpck')):
            uo = mpck()
        if uo is None:
            uo = json()
        uo_exists = uo.exists

        mtax_obj = None
        parent_name = f"{self.parent.prefix}.{self.parent.infix}.{self.parent.suffix}"
        if uo_exists:
            try:
                if parent_name != f"{self!s}":
                    mtax_obj = uo.model(
                        read=True
                    )["mtax"][f"{parent_name}"][f"{self!s}"]
                else:
                    mtax_obj = uo.model(
                        read=True
                    )["mtax"][f"{self!s}"]
            except (KeyError, JSONDecodeError):
                return 'Notfound'
        if mtax_obj is not None:
            return mtax_obj
        return 'Notfound'



import shutil

from montyalex.fs_tools import pathexists


ERROR = '[red]Error![/]'

def remove_cache__dir(formatted_directory: str | bytes, silent: bool = False) -> bool:
    if pathexists(formatted_directory):
        try:
            shutil.rmtree(formatted_directory)
            return True
        except (OSError, NotADirectoryError, IsADirectoryError):
            if not silent:
                print(
                    f'{ERROR}, Removing {formatted_directory!r} failed in the directory given')
            return False
    else:
        if not silent:
            print(
                f'Warning, Attempted to remove {formatted_directory!r}, ' + "but it didn't exist")
        return False



from functools import cache

from montyalex.typer_tools import Option, Typer
from .mtax_cache import MtaxCache
from .timecache_app import time_


cache_: Typer = Typer(
    name='cache',
    add_help_option=False)
cache_.add_typer(time_)

@cache_.command(name='add', add_help_option=False)
def add_(
    *,
    key: str,
    value: str,
    silent: bool = Option(
        False, '-s', show_default='-!s', help='Silence extreneous messages in console'),
    timezone: str = 'Etc/Greenwich'):
    """Add a key-value pair in the cache"""
    mtax: MtaxCache = MtaxCache(timezone)
    mtax.add_item((key, value), silent)

@cache
@cache_.command(name='incr', add_help_option=False)
def incr_(
    *,
    key: str,
    timezone: str = 'Etc/Greenwich',
    silent: bool = Option(
        False, '-s', show_default='-!s', help='Silence extreneous messages in console')):
    """Increment an item count in the cache"""
    mtax: MtaxCache = MtaxCache(timezone)
    return mtax.incr_item(key, silent)

@cache_.command(name='set', add_help_option=False)
def set_(
    *,
    key: str,
    value: str,
    timezone: str = 'Etc/Greenwich',
    silent: bool = Option(
        False, '-s', show_default='-!s', help='Silence extreneous messages in console')):
    """Sets a key-value pair in the cache"""
    mtax: MtaxCache = MtaxCache(timezone)
    mtax.set_item((key, value), silent)

@cache
@cache_.command(name='get', add_help_option=False)
def get_(
    *,
    key: str,
    silent: bool = Option(
        False, '-s', show_default='-!s', help='Silence extreneous messages in console')):
    """Gets a key-value pair from the cache if the key exists"""
    mtax: MtaxCache = MtaxCache()
    return mtax.get_item(key, silent)

@cache
@cache_.command(name='decr', add_help_option=False)
def decr_(
    *,
    key: str,
    timezone: str = 'Etc/Greenwich',
    silent: bool = Option(
        False, '-s', show_default='-!s', help='Silence extreneous messages in console')):
    """Decrement an item count in the cache"""
    mtax: MtaxCache = MtaxCache(timezone)
    return mtax.decr_item(key, silent)

@cache_.command(name='clear', add_help_option=False)
def clear_(
    *,
    timezone: str = 'Etc/Greenwich',
    silent: bool = Option(
    False, '-s', show_default='-!s', help='Silence extreneous messages in console')):
    """Clear all the key-value pairs in the cache"""
    mtax: MtaxCache = MtaxCache(timezone)
    mtax.clear(silent)

@cache_.command(name='reset', add_help_option=False)
def reset_(
    *,
    timezone: str = 'Etc/Greenwich',
    silent: bool = Option(
    False, '-s', show_default='-!s', help='Silence extreneous messages in console')):
    """Reset stuck key-value pairs in the cache"""
    mtax: MtaxCache = MtaxCache(timezone)
    mtax.reset(silent)

@cache_.command(name='remove', add_help_option=False)
def remove_(
    *,
    key: str,
    timezone: str = 'Etc/Greenwich',
    silent: bool = Option(
        False, '-s', show_default='-!s', help='Silence extreneous messages in console')):
    """Remove the matching key from the cache"""
    mtax: MtaxCache = MtaxCache(timezone)
    mtax.remove(key, silent)

@cache_.command(name='list', add_help_option=False)
def list_(
    *,
    timezone: str = 'Etc/Greenwich',
    silent: bool = Option(
        False, '-s', show_default='-!s', help='Silence extreneous messages in console')):
    """List the key-value pairs stored in the cache"""
    mtax: MtaxCache = MtaxCache(timezone)
    mtax.list_k_v_pairs(silent)

@cache_.command(name='info', add_help_option=False)
def info_(
    *,
    timezone: str = 'Etc/Greenwich',
    list_items: bool = Option(
        False, '-list', help='List the key-value pairs stored in the cache (list command)'),
    silent: bool = Option(
        False, '-s', show_default='-!s', help='Silence extreneous messages in console')):
    """Shows various info and current size of caches"""
    mtax: MtaxCache = MtaxCache(timezone)
    mtax.info(list_items, silent)



from enum import Enum
from rich.table import Table

from montyalex.console_tools import richconsole
from montyalex.fs_tools import cancel
from montyalex.time_tools import func_time
from montyalex.typing_tools import Any, Generator


print = richconsole.print


class BaseName(Enum):
    BINARY = 2
    TERNARY = 3
    QUATERNARY = 4
    QUINARY = 5
    SENARY = 6
    SEPTENARY = 7
    OCTAL = 8
    NONARY = 9
    DECIMAL = 10
    UNDECIMAL = 11
    DUODECIMAL = 12
    TRIDECIMAL = 13
    TETRADECIMAL = 14
    PENTADECIMAL = 15
    HEXADECIMAL = 16
    VIGESIMAL = 20
    DUOVIGESIMAL = 22
    TETRAVIGESIMAL = 24
    TRIGESIMAL = 30
    DUOTRIGESIMAL = 32
    TETRAGESIMAL = 40
    QUADRAGESIMAL = 48
    SEXAGESIMAL = 60
    TETRASEXAGESIMAL = 64
    CENTESIMAL = 100

    @classmethod
    def get_name(cls, base: int) -> str:
        for name, value in cls.__members__.items():
            if value.value == base:
                return name.lower()
        return "unknown"

class DigitPreset:
    def __init__(self, digit: int, preset: str) -> None:
        self.digit = digit
        self.preset = preset

    def __str__(self) -> str:
        return self.preset

    def generate_base_dict(self, base: int) -> dict[str, Any]:
        chars = (
            "0123456789"
            "abcdefghijklmnopqrstuvwxyz"
            "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
            "!@#$%^&*()-=_+[]{}|;:'\",.<>?\\/"
            "±§µ¢£€¥¤©®™✓✗♥♦♣♠♪♫◊")
        max_value = base ** self.digit
        base_dict = {}
        for i in range(max_value):
            val = ""
            num = i
            while num > 0:
                val = chars[num % base] + val
                num //= base
            base_dict[val.zfill(self.digit)] = i
        return base_dict

    def __call__(self, base: int, zero_exclusive=True) -> dict[str, Any]:
        base_dict = self.generate_base_dict(base)
        if zero_exclusive:
            base_dict = {k: v for k, v in base_dict.items() if k != ("0" * self.digit)}
        return base_dict

    @staticmethod
    def display_range_with_decimal(base_dict: dict) -> None:
        for base_value, decimal_value in base_dict.items():
            print(f"{base_value} = {decimal_value}")

    @staticmethod
    def range_length(
        dict_length: int, range_name: str, show_wz_range: bool = False) -> tuple[str, str, str]:
        if show_wz_range:
            return "(%%)", f"{range_name} size", f"[blue1]±{dict_length}[/]"
        return "(++)", f"{range_name} size", f"[cyan]±{dict_length}[/]"

    def _get_base_dict(self, base: int, show_wz_range: bool = False) -> dict:
        base_dict = self(base, not show_wz_range)
        return base_dict

    def base_dict_length(self, base: int, show_wz_range: bool = False) -> int:
        length = base ** self.digit
        if show_wz_range:
            length += 1
        return length

    def set_of_base(self, base: int, show_wz_range: bool = False) -> set[str]:
        base_dict = self._get_base_dict(base, show_wz_range)
        return set(base_dict.keys())

    def ul_of_base(self, base: int, show_wz_range: bool = False) -> list[str]:
        return list(self.set_of_base(base, show_wz_range))

    def ol_of_base(self, base: int, show_wz_range: bool = False) -> list[str]:
        return sorted(self.ul_of_base(base, show_wz_range))

class DoubleDigitPreset(DigitPreset):
    def __init__(self) -> None:
        super().__init__(2, 'Double')

class TripleDigitPreset(DigitPreset):
    def __init__(self) -> None:
        super().__init__(3, 'Triple')

class QuadrupleDigitPreset(DigitPreset):
    def __init__(self) -> None:
        super().__init__(4, 'Quadruple')

class QuintupleDigitPreset(DigitPreset):
    def __init__(self) -> None:
        super().__init__(5, 'Quintuple')

class SextupleDigitPreset(DigitPreset):
    def __init__(self) -> None:
        super().__init__(6, 'Sextuple')

class SeptupleDigitPreset(DigitPreset):
    def __init__(self) -> None:
        super().__init__(7, 'Septuple')

class OctupleDigitPreset(DigitPreset):
    def __init__(self) -> None:
        super().__init__(8, 'Octuple')

class NonupleDigitPreset(DigitPreset):
    def __init__(self) -> None:
        super().__init__(9, 'Nonuple')

class DecupleDigitPreset(DigitPreset):
    def __init__(self) -> None:
        super().__init__(10, 'Decuple')


class Preset:
    digit_default_dict = {
        "single": DigitPreset(1, 'Single'),
        "double": DoubleDigitPreset(),
        "triple": TripleDigitPreset(),
        "quadruple": QuadrupleDigitPreset(),
        "quintuple": QuintupleDigitPreset(),
        "sextuple": SextupleDigitPreset(),
        "septuple": SeptupleDigitPreset(),
        "octuple": OctupleDigitPreset(),
        "nonuple": NonupleDigitPreset(),
        "decuple": DecupleDigitPreset()
    }

    def __init__(
        self,
        kind: str,
        *,
        name: str = None,
        formatted: str | bool = False,
        numbase: int = 10) -> None:
        self.kind = kind

        self.name = name
        self.format = formatted
        self.numbase = numbase
        self.preset: DigitPreset | object = None

        if self.kind == 'digit':
            if self.name is None:
                self.name = 'triple'
            self.preset = self.digit_default_dict[self.name]

    def __len__(self) -> int:
        if self.kind == 'digit':
            return self.preset.base_dict_length(self.numbase, False)
        return 0

    def _get_base_dict(self, base: int, show_wz_range: bool = False) -> dict:
        base_dict = self.digit_default_dict[self.name](base, not show_wz_range)
        return base_dict

    def _table_items_generator(self, show_wz_range: bool) -> Generator[tuple[dict, int], Any, None]:
        base_dict = self._get_base_dict(self.numbase, show_wz_range)
        yield base_dict

    @func_time('took; to generate')
    def table(self, show_wz_range: bool = False) -> Table:
        target_base_name = BaseName.get_name(self.numbase)
        rows = []
        for base_dict in self._table_items_generator(show_wz_range):
            rows.append(self.preset.range_length(len(base_dict), target_base_name, show_wz_range))
        t = Table(
            show_header=False,
            pad_edge=False,
            highlight=True,
            title=f'{self.name.capitalize()} Digit Range',
            title_style="bold turquoise2",
            title_justify="left"
        )
        t.add_column("1", no_wrap=True)
        t.add_column("2", no_wrap=True)
        t.add_column("3", justify="right", no_wrap=True)
        for row in rows:
            t.add_row(*row)
        return t

binary_single = Preset('digit', name='single', numbase=2)
binary_double = Preset('digit', name='double', numbase=2)
binary_triple = Preset('digit', name='triple', numbase=2)
binary_quadruple = Preset('digit', name='quadruple', numbase=2)
binary_quintuple = Preset('digit', name='quintuple', numbase=2)
binary_sextuple = Preset('digit', name='sextuple', numbase=2)
binary_septuple = Preset('digit', name='septuple', numbase=2)
binary_octuple = Preset('digit', name='octuple', numbase=2)
binary_nonuple = Preset('digit', name='nonuple', numbase=2)
binary_decuple = Preset('digit', name='decuple', numbase=2)

ternary_single = Preset('digit', name='single', numbase=3)
ternary_double = Preset('digit', name='double', numbase=3)
ternary_triple = Preset('digit', name='triple', numbase=3)
ternary_quadruple = Preset('digit', name='quadruple', numbase=3)
ternary_quintuple = Preset('digit', name='quintuple', numbase=3)
ternary_sextuple = Preset('digit', name='sextuple', numbase=3)
ternary_septuple = Preset('digit', name='septuple', numbase=3)
ternary_octuple = Preset('digit', name='octuple', numbase=3)
ternary_nonuple = Preset('digit', name='nonuple', numbase=3)
ternary_decuple = Preset('digit', name='decuple', numbase=3)

quaternary_single = Preset('digit', name='single', numbase=4)
quaternary_double = Preset('digit', name='double', numbase=4)
quaternary_triple = Preset('digit', name='triple', numbase=4)
quaternary_quadruple = Preset('digit', name='quadruple', numbase=4)
quaternary_quintuple = Preset('digit', name='quintuple', numbase=4)
quaternary_sextuple = Preset('digit', name='sextuple', numbase=4)
quaternary_septuple = Preset('digit', name='septuple', numbase=4)
quaternary_octuple = Preset('digit', name='octuple', numbase=4)
quaternary_nonuple = Preset('digit', name='nonuple', numbase=4)
quaternary_decuple = Preset('digit', name='decuple', numbase=4)

quinary_single = Preset('digit', name='single', numbase=5)
quinary_double = Preset('digit', name='double', numbase=5)
quinary_triple = Preset('digit', name='triple', numbase=5)
quinary_quadruple = Preset('digit', name='quadruple', numbase=5)
quinary_quintuple = Preset('digit', name='quintuple', numbase=5)
quinary_sextuple = Preset('digit', name='sextuple', numbase=5)
quinary_septuple = Preset('digit', name='septuple', numbase=5)
quinary_octuple = Preset('digit', name='octuple', numbase=5)
quinary_nonuple = Preset('digit', name='nonuple', numbase=5)

senary_single = Preset('digit', name='single', numbase=6)
senary_double = Preset('digit', name='double', numbase=6)
senary_triple = Preset('digit', name='triple', numbase=6)
senary_quadruple = Preset('digit', name='quadruple', numbase=6)
senary_quintuple = Preset('digit', name='quintuple', numbase=6)
senary_sextuple = Preset('digit', name='sextuple', numbase=6)
senary_septuple = Preset('digit', name='septuple', numbase=6)
senary_octuple = Preset('digit', name='octuple', numbase=6)
senary_nonuple = Preset('digit', name='nonuple', numbase=6)

septenary_single = Preset('digit', name='single', numbase=7)
septenary_double = Preset('digit', name='double', numbase=7)
septenary_triple = Preset('digit', name='triple', numbase=7)
septenary_quadruple = Preset('digit', name='quadruple', numbase=7)
septenary_quintuple = Preset('digit', name='quintuple', numbase=7)
septenary_sextuple = Preset('digit', name='sextuple', numbase=7)
septenary_septuple = Preset('digit', name='septuple', numbase=7)
septenary_octuple = Preset('digit', name='octuple', numbase=7)

octal_single = Preset('digit', name='single', numbase=8)
octal_double = Preset('digit', name='double', numbase=8)
octal_triple = Preset('digit', name='triple', numbase=8)
octal_quadruple = Preset('digit', name='quadruple', numbase=8)
octal_quintuple = Preset('digit', name='quintuple', numbase=8)
octal_sextuple = Preset('digit', name='sextuple', numbase=8)
octal_septuple = Preset('digit', name='septuple', numbase=8)

nonary_single = Preset('digit', name='single', numbase=9)
nonary_double = Preset('digit', name='double', numbase=9)
nonary_triple = Preset('digit', name='triple', numbase=9)
nonary_quadruple = Preset('digit', name='quadruple', numbase=9)
nonary_quintuple = Preset('digit', name='quintuple', numbase=9)
nonary_sextuple = Preset('digit', name='sextuple', numbase=9)
nonary_septuple = Preset('digit', name='septuple', numbase=9)

decimal_single = Preset('digit', name='single')
decimal_double = Preset('digit', name='double')
decimal_triple = Preset('digit', name='triple')
decimal_quadruple = Preset('digit', name='quadruple')
decimal_quintuple = Preset('digit', name='quintuple')
decimal_sextuple = Preset('digit', name='sextuple')
decimal_septuple = Preset('digit', name='septuple')
decimal_octuple = Preset('digit', name='octuple')
decimal_nonuple = Preset('digit', name='nonuple')
decimal_decuple = Preset('digit', name='decuple')

duodecimal_single = Preset('digit', name='single', numbase=12)
duodecimal_double = Preset('digit', name='double', numbase=12)
duodecimal_triple = Preset('digit', name='triple', numbase=12)
duodecimal_quadruple = Preset('digit', name='quadruple', numbase=12)
duodecimal_quintuple = Preset('digit', name='quintuple', numbase=12)
duodecimal_sextuple = Preset('digit', name='sextuple', numbase=12)

hexadecimal_single = Preset('digit', name='single', numbase=16)
hexadecimal_double = Preset('digit', name='double', numbase=16)
hexadecimal_triple = Preset('digit', name='triple', numbase=16)
hexadecimal_quadruple = Preset('digit', name='quadruple', numbase=16)
hexadecimal_quintuple = Preset('digit', name='quintuple', numbase=16)

base32_single = Preset('digit', name='single', numbase=32)
base32_double = Preset('digit', name='double', numbase=32)
base32_triple = Preset('digit', name='triple', numbase=32)
base32_quadruple = Preset('digit', name='quadruple', numbase=32)

base64_single = Preset('digit', name='single', numbase=64)
base64_double = Preset('digit', name='double', numbase=64)
base64_triple = Preset('digit', name='triple', numbase=64)

LPR: dict[str, Preset] = {
    'single': binary_single,
    'double': binary_double,
    'triple': binary_triple,
    'quadruple': binary_quadruple,
    'quintuple': binary_quintuple,
    'sextuple': binary_sextuple,
    'septuple': binary_septuple,
    'octuple': binary_octuple,
    'nonuple': binary_nonuple,
    'decuple': binary_decuple
}
GPR: dict[str, Preset] = {
    'single': base64_single,
    'double': base64_double,
    'triple': base64_triple,
    'quadruple': base32_quadruple,
    'quintuple': hexadecimal_quintuple,
    'sextuple': duodecimal_sextuple,
    'septuple': nonary_septuple,
    'octuple': septenary_octuple,
    'nonuple': senary_nonuple,
    'decuple': quaternary_decuple
}
MCR: dict[str, Preset] = {
    'single': decimal_single,
    'double': decimal_double,
    'triple': decimal_triple,
    'quadruple': decimal_quadruple,
    'quintuple': decimal_quintuple,
    'sextuple': decimal_sextuple,
    'septuple': decimal_septuple,
    'octuple': decimal_octuple,
    'nonuple': decimal_nonuple,
    'decuple': decimal_decuple
}
# ---------------------------------------------------------------------------
# | Example Usage
# ---------------------------------------------------------------------------
if __name__ == '__main__':
    def main():
        digits = (
            '(single|double|triple|quadruple|quintuple|'
            'sextuple|septuple|octuple|nonuple|decuple)')
        digits_input = input(
            f'How many digits should be used? {digits}: ').lower().strip()
        base = input('Which base should be used? (10:default|2-64): ').strip()
        if base == '' or not base.isnumeric():
            base = 10
        else:
            base = int(base)
            if base > 64 or base < 2:
                print(f'Base {base} not allowed')
                cancel()

        double_up_input = input(
            'Do you want to double the folder name? (a=aa, y|N): ').lower().strip()
        zeroith_input = input('Do you want to include a zeroith folder? (y|N): ').lower().strip()
        _preset = MCR[f'{digits_input}'].preset.ol_of_base(base, bool(zeroith_input == 'y'))
        for b in _preset:
            print(b)
            if double_up_input == 'y':
                print(f'{b}_{b}')
        if double_up_input == 'y':
            print(f'len({len(_preset) * 2})')
        else:
            print(f'len({len(_preset)})')



from montyalex.fs_tools import current_working_dir, joinpaths, pathexists, mkdirs, rmfile
from montyalex.object_tools import singleton, pdumps, ploads


@singleton(get_identifier=lambda name: name)
class FileSystemTree:
    def __init__(self, name: str):
        self.name: str = name
        self.age = 0
        self.template: str = None
        self.nodes = [0, 0, 0]
        self._path: str = joinpaths(
            current_working_dir,
            '.mtax',
            '.pkl',
            f'tree-{self.name}')
        self.exists: bool = pathexists(self._path)
    def __str__(self) -> str:
        return f"{self.__dict__}"
    def progress(self):
        self.age += 1
        self.write()
    def write(self) -> None:
        mkdirs(self._path.removesuffix(f'tree-{self.name}'), exist_ok=True)
        with open(self._path, 'wb') as tree_pkl:
            tree_pkl.write(pdumps(self.__dict__))
    def setup(self, *, template: str = None):
        if template:
            self.template = template
    def scan(self):
        if self.exists:
            with open(self._path, 'rb') as tree_pkl:
                return ploads(tree_pkl.read())
        return self.__dict__
    def delete(self):
        if self.exists:
            rmfile(self._path)


# ---------------------------------------------------------------------------
# | Example Usage
# ---------------------------------------------------------------------------
if __name__ == '__main__':
    tree1 = FileSystemTree('name')
    print(tree1)
    print(tree1.scan())
